from django.contrib.auth import get_user_model, logout
from django.contrib.auth.forms import UserCreationForm
from django.conf import settings
from django import forms
from django.forms.utils import ErrorDict
from django.test import override_settings
from django.core.cache import cache
from django.core.exceptions import SuspiciousOperation
from django.contrib.auth.hashers import make_password, check_password

from ..utils import SecurePickling, send_template_email


UserModel = get_user_model()
secure_pickle = SecurePickling()

override_setting_decorator = override_settings(PASSWORD_HASHERS=settings.REGISTRATION_HASHERS)
hash_data = override_setting_decorator(make_password)
check_data = override_setting_decorator(check_password)


class RegistrationForm(UserCreationForm):
    error_messages = {
        'password_mismatch': "The two password fields didn't match.",
    }

    first_name = forms.CharField(max_length=30, required=False, help_text='Optional.')
    last_name = forms.CharField(max_length=30, required=False, help_text='Optional.')
    email = forms.EmailField(max_length=254, help_text='Required. Inform a valid email address.')
    if settings.SEND_CONFIRMATION_EMAIL:
        uid = forms.CharField(max_length=256, required=False)
    if settings.ENABLE_AGREEMENT_TERMS:
        agreement = forms.BooleanField(required=False)

    class Meta(UserCreationForm.Meta):
        model = UserModel
        fields = [
            'username',
            'first_name',
            'last_name',
            'email',
            'password1',
            'password2',
        ]

    def build_confirmation_url(self, uid):
        # pylint: disable=no-member
        return self.request.build_absolute_uri(f'?uid={uid}')

    def get_email_context(self):
        return {}

    def register_with_confirmation(self, commit):
        cache_key = self.cleaned_data.get('uid', None)

        if cache_key in (None, ''):
            cache_key = hash_data(self.cleaned_data['email'])
            self.cleaned_data['uid'] = cache_key

            secured_data = secure_pickle.dumps(self.cleaned_data)
            cache.set(cache_key, secured_data)
            context_data = self.get_email_context()
            context_data['action_url'] = self.build_confirmation_url(cache_key)
            if context_data.get('application_name', None) is None:
                context_data['application_name'] = settings.PROJECT_GUI_NAME

            send_template_email(
                subject='Registration Confirmation.',
                template_name='registration/confirm_email.html',
                email=self.cleaned_data['email'],
                context_data=context_data
            )
            return super().save(commit=False)

        if not check_data(self.cleaned_data['email'], cache_key):
            raise SuspiciousOperation('Invalid registration email send.')

        return super().save(commit)

    def save(self, commit=True):
        if settings.SEND_CONFIRMATION_EMAIL:
            return self.register_with_confirmation(commit)
        return super(RegistrationForm, self).save(commit)

    def _clean_fields(self):
        super()._clean_fields()
        if settings.SEND_CONFIRMATION_EMAIL:
            uid = self.cleaned_data.get('uid', None)
            if self.errors and uid not in (None, ''):
                self.cleaned_data.update(secure_pickle.loads(cache.get(uid)))
                self._errors = ErrorDict()

    # method clean is not needed?
    def clean(self):
        super().clean()
        if settings.ENABLE_AGREEMENT_TERMS:
            agreement = self.cleaned_data.get('agreement', None)
            if not agreement:
                self.add_error('agreement', 'To continue, need to accept the terms agreement.')


class TwoFaForm(forms.Form):
    pin = forms.CharField(required=True)

    def __init__(self, *args, **kwargs):
        self.request = kwargs.pop('request', None)
        super(TwoFaForm, self).__init__(*args, **kwargs)

    def clean(self):
        pin = self.cleaned_data.get('pin')

        if self.request.user.twofa.verify(pin):
            self.request.session['2fa'] = True
        else:
            self.add_error('pin', 'Invalid authentication code')
            attempts = self.request.session.get('2fa_attempts', 0) + 1
            if attempts >= settings.MAX_TFA_ATTEMPTS:
                logout(self.request)
            else:
                self.request.session['2fa_attempts'] = attempts

        return super().clean()

    def get_user(self):
        return self.request.user



# flake8: noqa: E501
TRANSLATION = {
    "new": '新的',
    "Create": '建立',
    "add": '增加',
    "Edit": '校对',
    "edit": '校对',
    "Save": '储存',
    "Remove": '删除',
    "Reload": '重启',
    "send": '发送',
    "select": '选择',
    "filters": '过滤',
    "apply": '应用',
    "Cancel": '取消',
    "field": '范围',
    "actions": '活动',
    "sublinks": '输入链接',
    "name": '种类',
    "type": '类型',
    "owner": '业主',
    "notes": '标记',
    "copy": '复制',
    "error": '错误',
    "home": '主要的',
    "login": '进入',
    "logout": '退出',
    "profile": '轮廓',
    "settings": '设置',
    "Settings": '设置',
    "version": '版本',
    "cache": '隐藏',
    "documentation": '文件',
    "request": '询问',
    "repository": '复位器',
    "app info": '信息',
    "status": '状态',
    "list is empty": '空白表格',
    "child instances": '输入对象',
    "enter value": '输入意义',
    "More info": '输入价值',
    "more info": '输入价值',
    "search by": '更详尽',
    'system': '范围查找',
    'users': '系统',
    'Is active': '活跃度',
    'Is staff': '减速装置',
    'Field "<b>{0}</b>" is empty.': '范围"<b>{0}</b>" 是空的',
    'Field "<b>{0}</b>" is required.': '范围"<b>{0}</b>" 必须完成',
    'Field "<b>{0}</b>" is too short.<br> Field length should not be shorter, than {1}.': '长度范围 "<b>{0}</b>" 太少.<br>长度范围应不少于{1}.',
    'Field "<b>{0}</b>" is too long. <br> Field length should not be longer, than {1}.': '长度单位"<b>{0}</b>" 太大<br>长度单位 应不大于{1}.',
    'Field "<b>{0}</b>" is too small.<br> Field should not be smaller, than {1}.': '取值范围 "<b>{0}</b>" 太少.<br> 取值范围应不少于 {1}.',
    'Field "<b>{0}</b>" is too big.<br> Field should not be bigger, than {1}.': '取值范围 "<b>{0}</b>" 太大.<br> 取值范围应不大于{1}.',
    '<b>{0} </b> value is not valid for <b>{1}</b> field.': '不能在 <b>{1}</b>取值为<b>{0}</b> .',
    '<b>"{0}"</b> field should be written in <b>"example@mail.com"</b> format.': '应在 <b>"example@mail.com"</b>.范围内完成<b>"{0}"</b>',
    'Child "<b>{0}</b>" instance was successfully added to parent list.': '子对象“ <b> {0} </ b>”已成功添加到母列表中.',
    'New "<b>{0}</b>" instance was successfully created.': '新对象“ <b> {0} </ b>”已成功创建.',
    '"<b>{0}</b>" {1} was successfully removed.': '"<b>{0}</b>" {1} 已成功删除."',
    'Changes in "<b>{0}</b>" {1} were successfully saved.': '对象变化{1}“ <b> {0} </ b>”的更改已成功保存.',
    'Action "<b>{0}</b>" was successfully executed on "<b>{1}</b>" instance.': '在对象“ <b> {1} </ b>”上成功启动了操作“ <b> {0} </ b>”。',
    'Some error occurred during adding of child "<b>{0}</b>" instance to parent list.<br> Error details: {1}': '将对象“ <b> {0} </ b>”添加到目级列表时发生错误。<br>更多详细信息：{1}',
    'Some error occurred during new "<b>{0}</b>" instance creation.<br> Error details: {1}': '创建新对象“ <b> {0} </ b>”时发生错误<br>更多详细信息：{1}',
    'Some error occurred during remove process of "<b>{0}</b>" {1}.<br> Error details: {2}': '删除{1}“ <b> {0} </ b>”时发生错误<br>更多详细信息：{2}',
    'Some error occurred during saving process of "<b>{0}</b>" {1}.<br> Error details: {2}': '保存{1}“ <b> {0} </ b>”时发生错误。<br>更多详细信息：{2}',
    'Some error occurred during "<b>{0}</b>" action execution on {1}.<br> Error details: {2}': '在{1}上执行操作“ <b> {0} </ b>”时发生错误<br>更多详细信息：{2}',
    'customize application styles': '自定义应用程序颜色样式',
    'skin': '主题',
    'skin settings': '主题设置',
    'active menu background': '活动菜单项的背景',
    'active menu color': '活动菜单项',
    'body background': '页面背景',
    'top navigation background': '顶部导航框的背景',
    'top navigation border': '顶部导航块的边框',
    'left sidebar background': '左侧边栏背景',
    'left sidebar border': '左侧的侧边栏',
    'customizer sidebar background': '侧边栏的背景',
    'breadcrumbs background': '面包屑背景',
    'buttons default background': '标准按钮背景',
    'buttons default text': '标准按钮背景',
    'buttons default border': '标准按钮边框',
    'buttons primary background': '主按钮背景',
    'buttons primary text': '主按钮文字',
    'buttons primary border': '主按钮边框',
    'buttons danger background': '背景危险按钮',
    'buttons danger text': '文字危险按钮',
    'buttons danger border': '边缘危险按钮',
    'buttons warning background': '背景提醒按钮',
    'buttons warning text': '警告按钮文字',
    'buttons warning border': '警告按钮边框',
    'links': '链接',
    'links hover': '悬停链接',
    'links revers': '链接暗色背景',
    'links hover revers': '悬停在黑暗背景上的链接',
    'text color': '文章',
    'ico default color': '标准图标颜色',
    'text header color': '标题',
    'background default color': '背景标准色',
    'card header background': '“卡片”部分顶部背景',
    'card body background': '“卡片”部分主题背景',
    'card footer background': '“卡片”部分底部背景',
    'card color': '插件中的文本“卡片”',
    'labels': '标签',
    'help content': '支持内容',
    'help text color': '支持文章',
    'table line hover bg color': '悬停表行背景',
    'table border color': '表格边框',
    'table line selected bg color': '所选表格行的背景',
    'background active color': '活动元素背景',
    'background passive color': '非活动元素背景',
    'text hover color': '悬停时的文字颜色',
    'boolean true color': '布尔值的颜色，值为“ true”',
    'boolean false color': '值为“ false”的布尔变量的颜色',
    'modal background color': '模态窗口背景',
    'api sections background color': 'api中的背景部分',
    'api sections border color': 'api中的边框部分',
    'code highlighting color 1': '代码突出显示-颜色1',
    'code highlighting color 2': '代码突出显示-颜色2',
    'code highlighting color 3': '代码突出显示-颜色3',
    'code highlighting color 4': '代码突出显示-颜色4',
    'code highlighting color 5': '代码突出显示-颜色5',
    'code highlighting color 6': '代码突出显示-颜色6',
    'code highlighting color 7': '代码突出显示-颜色7',
    'custom css': '自定义CSS',
    'reset skin settings to default': '恢复默认设置',
    'save skin': '储存主题',
    'Skin settings were successfully reset to default.': '主题设置已成功恢复为默认值',
    'Skin settings were successfully saved': '主题设置已成功保存',
    'actions on': '行动结束',
    'on item': '元素| 元素| 元素| 元素',
    'key': '要素',
    'value': '意义',
    'new owner': '新业主',
    'detail': '细节',
    'description': '描述',
    'id__not': '排除编号',
    'username': '使用者',
    'is active': '活跃度',
    'user': '使用者',
    'yes': '是',
    'no': '否',
    'is staff': '管理',
    'first_name': '名字',
    'last_name': '姓',
    'last name': '姓',
    'password': '密码',
    'Password': '密码',
    'Change password': '变换密码',
    'Change_password': '变换密码',
    'generate password': '形成密码',
    'repeat password': '重复密码',
    'Repeat password': '重复密码',
    'Old password': '旧密码',
    'New password': '新密码',
    'Confirm new password': '确认密码',
    'username__not': '排除用户名',
    'name__not': '排除名称',
    'kind': '类型',
    'mode': '制度',
    'options': '选项',
    'height': '高度',
    'width': '宽度',
    'information': '信息',
    'date': '日期',
    'file n selected': '未选择文件| 已选择1个文件| 已选择{n}个文件| 已选择{n}个文件',
    'image n selected': '未选择图像 已选择1张图片| 已选择{n}张图片| 已选择{n}张图片',
    'Ordering': '排列顺序',
    'is_active': '活跃度',
    'Username': '使用名',
    'first name': '名字',
    'First name': '名字',
    'Last name': '姓',
    'boolean value meaning status of user.': '指示用户状态的布尔值。',
    'users first name.': '用户名',
    'users last name.': '用户姓',
    'users e-mail value.': '用户邮箱',
    'which field to use when ordering the results.': '结果排序依据的字段',
    'a unique integer value (or comma separated list) identifying this instance.': '标识此对象的唯一数值（或用逗号分隔的序列）识别此对象',
    'a name string value (or comma separated list) of instance.': '对象名称-字符串值（或它们的顺序，用逗号分隔）',
    'page matching current url was not found': '找不到与指定网址匹配的页面',
    'sign in to start your session': '登录开始会议',
    'sign in': '进入',
    'sign up': '请注册',
    'forgot password': '忘记密码',
    'confirm password': '确认密码',
    'register': '注册',
    'fields with * is required': '标有*为必填项',
    'Forgotten your password? Enter your email address below, and we\'ll email instructions for setting a new one.': '忘记密码了吗？ 输入您的地址，我们将向您发送恢复指示',
    'reset my password': '问密码',
    'We\'ve emailed you instructions for setting your password, if an account exists with the email you entered. You should receive them shortly.': '我们已向您发送了一封电子邮件，其中包含有关重设密码的说明',
    'If you don\'t receive an email, please make sure you\'ve entered the address you registered with, and check your spam folder.': '如果您尚未收到电子邮件，请确认输入正确的地址并检查垃圾邮件文',
    'The password reset link was invalid, possibly because it has already been used. Please request a new password reset.': '无效的密码恢复链接，可能已被使用。 请再试一次。',
    'Please enter your new password twice so we can verify you typed it in correctly.': '请输入新密码',
    'change my password': '换密码',
    'Your password has been set.  You may go ahead and log in now.': '您的密码已成功设置。 您现在可以登录',
    'Please enter a correct %(username)s and password. Note that both fields may be case-sensitive.': '请输入正确的％（用户名）和密码。 注意，这些字段可能区分大小写',
    'or': '或者',
    'two-factor authentication code': '两因素授权码',
    'verify code': '确认密码',
    'Enter the code from the two-factor app on your mobile device. If you\'ve lost your device, you may enter one of your recovery codes.': '从您的移动设备上的两个应用程序中输入代码。 如果您丢失了设备，则可以输入恢复其中一个代码',
    'Invalid authentication code': '无效的授权码',
    'Scan the image with the two-factor authentication app on your phone. If you can’t use a barcode, enter this text code instead.': '使用手机上的两个身份验证应用程序扫描图像。 如果您不能使用条形码，请输入文本代码。',
    'Enter the six-digit code from the application': '从应用程序中输入六个字符的代码',
    'code': '代码',
    'recovery codes': '恢复代码',
    'Recovery codes are used to access your account in the event you cannot receive two-factor authentication codes.': '如果您无法获得两个身份验证码，则可以使用恢复码来访问您的帐户。',
    'Disabling Two-Factor Authentication (2FA) will decrease your account\'s security': '禁用两个身份验证会降低您帐户的安全性',
    'Enable': '打开',
    'Disable': '关闭',
    'Two factor authentication': '两个验证',
    'No {0} matches the given query.': '{0}没有一个与此请求匹配',
    'Password is not correct.': '密码不正确',
    'New passwords values are not equal.': '新密码值不相同',
    'This password is entirely numeric.': '该密码应使用完全数字',
    'Dark theme': '黑色主题',
    'Language': '语言',
    'unsupported image file format, expected ({0}), got {1}': '不支持的图像格式，支持（{0}），已收到{1}',
    'for some reason, this image file cannot be opened': '无法打开图像文件',
    'Invalid image {0}. Expected from {1} to {2}, got {3}': "{0}图片不正确。 预期从{1}到{2}，收到{3}",
    'imageValidationResolutionError': "必须至少为{min}且不超过{max}个像素",
    'Dark mode': '黑暗主题',


    # Проверить
    'Changes in settings are successfully saved. Please refresh the page.': '设置更改已成功保存。 请刷新页面。',
    'now': '眼下',
    'later': '之后',
    'i accept the ': '我接受 ',
    'terms of agreement': '协议条款',
    'Zoom': '飛漲',
    'Scale': '規模',
}



# flake8: noqa: E501
TRANSLATION = {
    'new': 'Mới',
    'Create': 'tạo',
    'add': 'thêm',
    'Edit': 'chỉnh sửa',
    'edit': 'Chỉnh sửa',
    'Save': 'lưu',
    'Remove': 'xóa',
    'Reload': 'làm mới',
    'send': 'gửi',
    'select': 'chọn',
    'filters': 'bộ lọc',
    'apply': 'áp dụng',
    'Cancel': 'Hủy',
    'field': 'tường',
    'actions': 'thao tác',
    'sublinks': 'đường dẫn',
    'name': 'tên',
    'type': 'loại',
    'owner': 'chủ sở hữu',
    'notes': 'ghi chú',
    'copy': 'sao chép',
    'error': 'lỗi',
    'home': 'trang chủ',
    'login': 'đăng nhập',
    'logout': 'đăng xuất',
    'profile': 'hồ sơ',
    'settings': 'thiết lập',
    'Settings': 'Thiết lập',
    'version': 'phiên bản',
    'cache': 'bộ nhớ đệm',
    'documentation': 'tài liệu',
    'request': 'yêu cầu',
    'repository': 'kho thông tin',
    'app info': 'thông tin',
    'status': 'trạng thái',
    'list is empty': 'danh sách trống',
    'child instances': 'thêm các mục lồng',
    'enter value': 'nhập giá trị',
    'More info': 'thêm thông tin',
    'more info': 'thêm thông tin',
    'search by': 'tìm kiếm theo',
    'system': 'Hệ thống',
    'users': 'Người dùng',
    'Is active': 'Hoạt động',
    'Is staff': 'Người điều hành',
    'Field "<b>{0}</b>" is empty.': 'Trường "<b> {0} </b>" trống không',
    'Field "<b>{0}</b>" is required.': 'Trường "<b> {0} </b>"  bắt buộc.',
    'Field "<b>{0}</b>" is too short.<br> Field length should not be shorter, than {1}.': 'Độ dài của trường "<b> {0} </b>" quá ngắn. <br> Độ dài của trường không được nhỏ hơn {1}.',
    'Field "<b>{0}</b>" is too long. <br> Field length should not be longer, than {1}.': 'Độ dài của trường "<b> {0} </b>" quá dài. <br> Trường không được dài hơn {1}.',
    'Field "<b>{0}</b>" is too small.<br> Field should not be smaller, than {1}.': 'Giá trị của trường "<b> {0} </b>" quá nhỏ. <br> Giá trị của trường không được nhỏ hơn {1}.',
    'Field "<b>{0}</b>" is too big.<br> Field should not be bigger, than {1}.': 'Giá trị của trường "<b> {0} </b>" quá lớn. <br> Giá trị của trường không được lớn hơn {1}.',
    '<b>{0} </b> value is not valid for <b>{1}</b> field.': 'Giá trị <b> {0} </b> không hợp lệ cho trường <b> {1} </b>.',
    '<b>"{0}"</b> field should be written in <b>"example@mail.com"</b> format.': 'Trường <b> "{0}" </b> phải làm như <b> "example@mail.com" </b>.',
    'Child "<b>{0}</b>" instance was successfully added to parent list.': 'Đối tượng con "<b> {0} </b>" đã được thêm thành công vào danh sách mẹ.',
    'New "<b>{0}</b>" instance was successfully created.': 'Đối tượng mới "<b> {0} </b>" đã được tạo thành công.',
    '"<b>{0}</b>" {1} was successfully removed.': '"<b> {0} </b>" {1} đã được xóa thành công."',
    'Changes in "<b>{0}</b>" {1} were successfully saved.': 'Thay đổi đối tượng {1} "<b> {0} </ b>" đã được lưu thành công.',
    'Action "<b>{0}</b>" was successfully executed on "<b>{1}</b>" instance.': 'Thao tác "<b> {0} </b>" đã được thực hiện thành công trên đối tượng "<b> {1} </b>".',
    'Some error occurred during adding of child "<b>{0}</b>" instance to parent list.<br> Error details: {1}': 'Khi thêm đối tượng con "<b> {0} </b>" vào danh sách mẹ đã xảy ra lỗi <br> thêm chi tiết: {1}',
    'Some error occurred during new "<b>{0}</b>" instance creation.<br> Error details: {1}': 'Khi tạo đối tượng mới "<b> {0} </b>". đã xảy ra lỗi <br> biết thêm chi tiết: {1}',
    'Some error occurred during remove process of "<b>{0}</b>" {1}.<br> Error details: {2}': 'Khi xóa {1} "<b> {0} </b>". đã xảy ra lỗi <br> thêm chi tiết: {2}',
    'Some error occurred during saving process of "<b>{0}</b>" {1}.<br> Error details: {2}': 'Khi lưu {1} "<b> {0} </b>". đã xảy ra lỗi <br> thêm chi tiết: {2}',
    'Some error occurred during "<b>{0}</b>" action execution on {1}.<br> Error details: {2}': 'Khi chạy tác vụ "<b> {0} </b>" {1}.  đã xảy ra lỗi <br> thêm chi tiết: {2}',
    'customize application styles': 'thiết lập màu sắc của phần mềm',
    'skin': 'màu',
    'skin settings': 'сài đặt màu',
    'active menu background': 'màu mục menu hoạt động',
    'active menu color': 'mục menu hoạt động',
    'body background': 'phông nền của trang',
    'top navigation background': 'phông nền của thanh điều hướng',
    'top navigation border': 'phác thảo của thanh điều hướng',
    'left sidebar background': 'phông nền của menu',
    'left sidebar border': 'phác thảo của menu',
    'customizer sidebar background': 'phông nền của menu này',
    'breadcrumbs background': 'phông nền của menu chính',
    'buttons default background': 'phông nền của nút mặc định',
    'buttons default text': 'chữ của nút mặc định',
    'buttons default border': 'phác thảo của nút mặc định',
    'buttons primary background': 'phông nền của nút chính',
    'buttons primary text': 'chữ của nút chính',
    'buttons primary border': 'phác thảo của nút chính',
    'buttons danger background': 'phông nền của nút "danger"',
    'buttons danger text': 'chữ của nút "danger"',
    'buttons danger border': 'phác thảo của nút "danger"',
    'buttons warning background': 'phông nền của nút "warning"',
    'buttons warning text': 'chữ của nút "warning"',
    'buttons warning border': 'phác thảo của nút "warning"',
    'links': 'đường dẫn',
    'links hover': 'đường dẫn khi di chuột',
    'links revers': 'đường dẫn trên nền tối',
    'links hover revers': 'đường dẫn trên nền tối khi di chuột',
    'text color': 'văn bản',
    'ico default color': 'màu bình thường',
    'text header color': 'tiêu đề',
    'background default color': 'màu nền mặc định',
    'card header background': 'phông nền của phần "card" trên',
    'card body background': 'phông nền của phần "card" chính',
    'card footer background': 'phông nền của phần "card" dưới',
    'card color': 'chữ của phần "card"',
    'labels': 'nhãn',
    'help content': 'nội dung phụ',
    'help text color': 'văn bản phụ',
    'table line hover bg color': 'phông nền của bảng màn hình',
    'table border color': 'bảng phác thảo',
    'table line selected bg color': 'phông nền của bảng đánh dấu',
    'background active color': 'phông nền của yếu tố hoạt động',
    'background passive color': 'phông nền của yếu tố không hoạt động',
    'text hover color': 'màu văn bản khi di chuột',
    'boolean true color': 'màu của từ "true"',
    'boolean false color': 'màu của từ  "false"',
    'modal background color': 'màu thuộc tính Modal',
    'api sections background color': 'màu Api',
    'api sections border color': 'phông nền Api',
    'code highlighting color 1': 'đèn nền - màu 1',
    'code highlighting color 2': 'đèn nền - màu 2',
    'code highlighting color 3': 'đèn nền - màu 3',
    'code highlighting color 4': 'đèn nền - màu 4',
    'code highlighting color 5': 'đèn nền - màu 5',
    'code highlighting color 6': 'đèn nền - màu 6',
    'code highlighting color 7': 'đèn nền - màu 7',
    'custom css': 'css tùy chỉnh',
    'reset skin settings to default': 'về cài đặt mặc định',
    'save skin': 'lưu màu',
    'Skin settings were successfully reset to default.': 'Cài đặt đã trở lại tiêu chuẩn',
    'Skin settings were successfully saved': 'Cài đặt đã lưu thành công',
    'actions on': 'thao tác trên',
    'on item': 'thành phần',
    'key': 'Chìa khóa',
    'value': 'Giá trị',
    'new owner': 'Chủ sở hữu mới',
    'detail': 'Chi tiết',
    'description': 'Miêu tả',
    'username': 'Tên người dùng',
    'is active': 'Hoạt động',
    'user': 'Người dùng',
    'yes': 'Có',
    'no': 'Không',
    'is staff': 'Admin',
    'first name': 'Tên',
    'last name': 'Họ',
    'password': 'Mật khẩu',
    'Password': 'Mật khẩu',
    'Change password': 'Đổi mật khẩu',
    'Change_password': 'Đổi mật khẩu',
    'generate password': 'Viết mật khẩu',
    'repeat password': 'Lập lại mật khẩu',
    'Repeat password': 'Lập lại mật khẩu',
    'Old password': 'Mật khẩu cũ',
    'New password': 'Mật khẩu mới',
    'Confirm new password': 'Xác nhận mật khẩu mới',
    'kind': 'Loại',
    'mode': 'Chế độ',
    'options': 'Tùy chọn',
    'height': 'chiều cao',
    'width': 'chiều rộng',
    'information': 'Thông tin',
    'Dark mode': 'Chủ đề tối',
    'date': 'Ngày',
    'file n selected': 'File chưa được chọn',
    'image n selected': 'Hình ảnh chưa được chọn',
    'Ordering': 'Thứ tự phân loại',
    'is_active': 'Hoạt động',
    'Username': 'Tên đăng nhập',
    'First name': 'Tên',
    'Last name': 'Họ',
    'boolean value meaning status of user.': 'giá trị boolean cho biết trạng thái của người dùng.',
    'users first name.': 'Tên người dùng',
    'users last name.': 'Họ người dùng',
    'users e-mail value.': 'E-mail người dùng',
    'which field to use when ordering the results.': 'trường để sắp xếp kết quả.',
    'a unique integer value (or comma separated list) identifying this instance.': 'giá trị số duy nhất',
    'a name string value (or comma separated list) of instance.': 'tên đối tượng',
    'page matching current url was not found': 'trang web theo URL này không thể tìm được',
    'sign in to start your session': 'đăng nhập để bắt đầu',
    'sign in': 'đăng nhập',
    'sign up': 'đăng ký',
    'forgot password': 'quên mật khẩu',
    'confirm password': 'xác nhận mật khẩu',
    'register': 'đăng ký',
    'fields with * is required': 'các mục có dấu * là bắt buộc',
    'Forgotten your password? Enter your email address below, and we\'ll email instructions for setting a new one.': 'Quên mật khẩu? Nhập E-mail để nhận hướng dẫn khôi phục mật khẩu',
    'reset my password': 'Hủy mật khẩu',
    'We\'ve emailed you instructions for setting your password, if an account exists with the email you entered. You should receive them shortly.': 'Hướng dẫn hủy mật khẩu đã được gửi vào e-mail của bạn.',
    'If you don\'t receive an email, please make sure you\'ve entered the address you registered with, and check your spam folder.': '"Nếu bạn chưa nhận được thư, hãy kiểm tra lại địa chỉ e-mail và mục Spam."',
    'The password reset link was invalid, possibly because it has already been used. Please request a new password reset.': '"Đường dẫn không đúng, có thể đã được sử dụng. Xin hãy thử lại lần nữa."',
    'Please enter your new password twice so we can verify you typed it in correctly.': 'Nhập mật khẩu mới',
    'change my password': 'Thay đổi mật khẩu',
    'Your password has been set.  You may go ahead and log in now.': 'Thay đổi mật khẩu thành công. Hãy đăng nhập.',
    'Please enter a correct %(username)s and password. Note that both fields may be case-sensitive.': 'Vui lòng nhập chính xác% (tên người dùng) s và mật khẩu',
    'or': 'hoặc',
    'two-factor authentication code': 'hai bước xác thực (2FA)',
    'verify code': 'xác nhận mã',
    'Field <b>{0}</b>" is too short.<br> Field length should not be shorter, than {1}.': 'Độ dài của trường "<b> {0} </b>" quá ngắn. <br> Độ dài của trường không được nhỏ hơn {1}.',
    'Field <b>{0}</b>" is too long. <br> Field length should not be longer, than {1}.': 'Độ dài của trường "<b> {0} </b>" quá dài. <br> Trường không được dài hơn {1}.',
    'Field <b>{0}</b>" is too small.<br> Field should not be smaller, than {1}.': 'Giá trị của trường "<b> {0} </b>" quá nhỏ. <br> Giá trị của trường không được nhỏ hơn {1}.',
    'Field <b>{0}</b>" is too big.<br> Field should not be bigger, than {1}.': 'Giá trị của trường "<b> {0} </b>" quá lớn. <br> Giá trị của trường không được lớn hơn {1}.',
    '"<b>{0}</b> {1} was successfully removed."': '"""<b> {0} </b>"" {1} đã được xóa thành công."',
    'Enter the code from the two-factor app on your mobile device. If you\'ve lost your device, you may enter one of your recovery codes.': '"Nhập mã từ ứng dụng sử dụng hai bước xác thực trên thiết bị di động của bạn.Nếu bạn mất thiết bị của mình, bạn có thể nhập một trong các mã khôi phục."',
    'Invalid authentication code': 'Mã xác thực không hợp lệ',
    'Scan the image with the two-factor authentication app on your phone. If you can’t use a barcode, enter this text code instead.': '"Quét hình ảnh bằng ứng dụng xác thực hai bước trên điện thoại của bạn. Nếu bạn không thể quét hình ảnh, hãy nhập mã văn bản."',
    'Enter the six-digit code from the application': 'Điền mã xác thực từ ứng dụng',
    'code': 'Mã',
    'recovery codes': 'Mã khôi phục',
    'Recovery codes are used to access your account in the event you cannot receive two-factor authentication codes.': 'Mã khôi phục được sử dụng để truy cập vào tài khoản của bạn nếu bạn không thể nhận được mã xác thực hai bước.',
    'Disabling Two-Factor Authentication (2FA) will decrease your account\'s security': 'mã xác thực hai bước.',
    'Enable': 'Bật',
    'Disable': 'Tắt',
    'Two factor authentication': 'Xác thực hai bước (2FA)',
    'No {0} matches the given query.': 'Không tìm thấy',
    'Password is not correct.': 'Sai mật khẩu',
    'New passwords values are not equal.': 'Mật khẩu mới không chính xác',
    'This password is entirely numeric.': 'Mật khẩu này chỉ có chữ số.',
    'Dark theme': 'Màn hình tối',
    'Language': 'Ngôn ngữ',
    'unsupported image file format, expected ({0}), got {1}': 'Định dạng hình ảnh không được hỗ trợ, Được hỗ trợ ({0}), đã nhận {1}',
    'for some reason, this image file cannot be opened': 'Không thể mở file hình ảnh',
    'Invalid image {0}. Expected from {1} to {2}, got {3}': 'Hình ảnh {0} không chính xác. Dự kiến ​​từ {1} đến {2}, đã nhận {3}',
    'imageValidationResolutionError': 'Phải có ít nhất {min} và không nhiều hơn {max} pixel.',

    # Проверить
    'Changes in settings are successfully saved. Please refresh the page.': 'Các thay đổi trong cài đặt đã được lưu thành công. Vui lòng làm mới trang.',
    'now': 'hiện nay',
    'later': 'một lát sau',
    'i accept the ': 'Tôi chấp nhận ',
    'terms of agreement': 'các điều khoản của thỏa thuận',
    'Zoom': 'Thu phóng',
    'Scale': 'Tỉ lệ',
}



import json

from django import template

register = template.Library()


@register.inclusion_tag('vst_inclusion_tags/b64_img_from_json_string.html')
def img_from_json(json_string, **kwargs):
    media_type = kwargs.get('media_type', 'image/png')
    tag_classes = kwargs.get('tag_classes', '')
    content_attribute = kwargs.get('content_attribute', 'content')

    try:
        content = json.loads(json_string)[content_attribute]
    except Exception:
        content = ''
    return {
        'media_type': media_type,
        'tag_classes': tag_classes,
        'content': content
    }



import base64
import typing as _t
import re
import warnings
from io import BytesIO
from pathlib import PurePosixPath

from rest_framework import serializers

try:
    from PIL import Image, UnidentifiedImageError

    has_pillow = True
except ImportError:  # nocv
    has_pillow = False


class RegularExpressionValidator:
    """
    Class for regular expression based validation

    :raises rest_framework.exceptions.ValidationError: in case value does not match regular expression
    """
    regexp: _t.Optional[_t.Pattern[_t.Text]] = None

    def __init__(self, regexp: _t.Optional[_t.Pattern] = None):
        """
        :param regexp: Compiled regular expression
        """
        self._regexp = regexp if regexp else self.regexp

    def __call__(self, value):
        if value and not self._regexp.match(value):
            raise serializers.ValidationError(self._error_msg())

    def _error_msg(self):
        return f'This field must match pattern {self._regexp.pattern}'


class UrlQueryStringValidator(RegularExpressionValidator):
    """Class for validation url query string, for example a=&b=1"""

    regexp = re.compile(r'^[^&?=].+=.*$')


class ImageValidator:
    """
    Base Image Validation class
    Validates image format
    Won't work if Pillow isn't installed
    :param extensions: Tuple or List of file extensions, that should pass the validation
    Raises rest_framework.exceptions.ValidationError: in case file extension are not in the list
    """
    warning_msg = 'Warning! Pillow is not installed, validation has not been done'
    default_extensions: _t.Union[_t.Tuple, _t.List] = [
        'bmp',
        'jpeg',
        'jpg',
        'png',
    ]

    def __init__(self, extensions: _t.Optional[_t.Union[_t.Tuple, _t.List]] = None, **kwargs):
        if extensions is not None:
            assert isinstance(extensions, (tuple, list)), "extensions must be list or tuple"
            self.extensions = extensions
        else:
            self.extensions = self.default_extensions
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __call__(self, value):
        if not self.has_pillow:
            warnings.warn(self.warning_msg, ImportWarning)
            return
        file_extension = PurePosixPath(value['name']).suffix
        if value and file_extension[1:] not in self.extensions:
            raise serializers.ValidationError(f'unsupported image file format,'
                                              f' expected ({",".join(self.extensions)}),'
                                              f' got {file_extension[1:]}')

    @property
    def has_pillow(self):
        """
        Check if Pillow is installed
        """
        return has_pillow


class ImageOpenValidator(ImageValidator):
    """
    Image validator that checks if image can be unpacked from b64 to PIL Image obj

    Raises rest_framework.exceptions.ValidationError if PIL throws error when trying to open image
    """
    error_msg = 'for some reason, this image file cannot be opened'

    def __call__(self, value):
        if not self.has_pillow:
            warnings.warn(self.warning_msg, ImportWarning)
            return
        super().__call__(value)
        try:
            self.img = Image.open(BytesIO(base64.b64decode(value['content'])))
        except UnidentifiedImageError:
            raise serializers.ValidationError(self.error_msg)


class ImageBaseSizeValidator(ImageOpenValidator):
    """
    Validates image size
    To use this class for validating image width/height, rewrite
    self.orientation to ('height',) or ('width',) or ('height', 'width')

    Raises rest_framework.exceptions.ValidationError: if not(min <= (height or width) <= max)
    """
    orientation: _t.Union[_t.Union[_t.Tuple[str], _t.Tuple[str, str]], _t.Tuple] = ()

    def __call__(self, value):
        if not self.has_pillow:
            warnings.warn(self.warning_msg, ImportWarning)
            return
        super().__call__(value)
        self.validate()

    def validate(self):
        for orientation in self.orientation:
            min_value = getattr(self, f'min_{orientation}', 1)
            max_value = getattr(self, f'max_{orientation}', float('inf'))
            value = getattr(self.img, orientation)
            if not (min_value <= value <= max_value):
                raise serializers.ValidationError(f'Invalid image {orientation}. Expected from {min_value}'
                                                  f' to {max_value}, got {value}')


class ImageHeightValidator(ImageBaseSizeValidator):
    """
    Wrapper for ImageBaseSizeValidator that validates only height

    :param min_height: minimal height of an image being validated
    :param max_height: maximal height of an image being validated
    """
    orientation = ('height',)


class ImageWidthValidator(ImageBaseSizeValidator):
    """
    Wrapper for ImageBaseSizeValidator that validates only width

    :param min_width: minimal width of an image being validated
    :param max_width: maximal width of an image being validated
    """
    orientation = ('width',)


class ImageResolutionValidator(ImageBaseSizeValidator):
    """
    Wrapper for ImageBaseSizeValidator that validates both height and width

    :param min_height: minimal height of an image being validated
    :param max_height: maximal height of an image being validated
    :param min_width: minimal width of an image being validated
    :param max_width: maximal width of an image being validated
    """
    orientation = ('height', 'width')



from smtplib import SMTPException

from celery import Celery
from celery.app.task import BaseTask
from celery.result import AsyncResult
from django.conf import settings

from .utils import import_class, send_template_email_handler

celery_app: Celery = import_class(
    settings.WORKER_OPTIONS['app'].replace(':', '.')  # type: ignore
)


class TaskClass(BaseTask):
    """
    Wrapper for Celery BaseTask class. Usage is same as Celery standard class, but you can execute task without
    creating instance with :meth:`TaskClass.do` method.
    Example:

        .. sourcecode:: python

            from vstutils.environment import get_celery_app
            from vstutils.tasks import TaskClass

            app = get_celery_app()

            class Foo(TaskClass):
                def run(*args, **kwargs):
                    return 'Foo task has been executed'

            app.tasks.register(Foo())

        Now you can call your task with various methods:
            - by executing Foo.do(*args, **kwargs)
            - get registered task instance like that - app.tasks['full_path.to.task.class.Foo']

        Also you can make you registered task periodic, by adding it to CELERY_BEAT_SCHEDULE in settings.py:

        .. sourcecode:: python

            CELERY_BEAT_SCHEDULE = {
                'foo-execute-every-month': {
                    'task': 'full_path.to.task.class.Foo',
                    'schedule': crontab(day_of_month=1),
                },
            }
    """

    # pylint: disable=abstract-method

    @property
    def name(self):
        """
        property for proper Celery task execution, needed for :meth:`TaskClass.do` method to work
        """
        return self.__name__

    @property
    def __name__(self):
        return f'{self.__class__.__module__}.{self.__class__.__name__}'

    @classmethod
    def do(cls, *args, **kwargs) -> AsyncResult:
        """
        Method which send signal to celery for start remote task execution.
        All arguments will passed to the task :meth:`TaskClass.run` method.
        """
        return cls().delay(*args, **kwargs)


class SendEmailMessage(TaskClass):
    """
    Task for sending bulk emails, all args and kwargs are passed to :func:`vstutils.utils.send_template_email_handler`.
    Usually you don't need to call this manually, this is called by :func:`vstutils.utils.send_template_email`.
    """

    def run(self, *args, **kwargs):
        try:
            send_template_email_handler(*args, **kwargs)
        except SMTPException as exc:
            raise self.retry(
                exc=exc,
                max_retries=settings.SEND_EMAIL_RETRIES,
                countdown=settings.SEND_MESSAGE_RETRY_DELAY
            )


celery_app.tasks.register(SendEmailMessage())  # type: ignore



"""
Additional serializer fields for generating OpenAPI and GUI.
"""
import logging
import typing as _t
import copy
import json
import functools
import base64

import orjson
from rest_framework.serializers import CharField, IntegerField, FloatField, ModelSerializer
from rest_framework.fields import empty, SkipField, get_error_detail, Field
from rest_framework.exceptions import ValidationError
from django.apps import apps
from django.db import models
from django.utils.functional import SimpleLazyObject, lazy
from django.core.exceptions import ValidationError as DjangoValidationError
from django.core.files.uploadedfile import SimpleUploadedFile
from django.db.models.fields.files import FieldFile

from ..models import fields as vst_model_fields
from ..utils import raise_context, get_if_lazy, raise_context_decorator_with_default

DependenceType = _t.Optional[_t.Dict[_t.Text, _t.Text]]
DEFAULT_NAMED_FILE_DATA = {"name": None, "content": None, 'mediaType': None}


class VSTCharField(CharField):
    """
    CharField (extends :class:`rest_framework.fields.CharField`).
    This field translate any json type to string for model.
    """

    __slots__ = ()

    def to_internal_value(self, data) -> _t.Text:
        with raise_context():
            if not isinstance(data, str):
                data = json.dumps(data)
        data = str(data)
        return super().to_internal_value(data)


class FileInStringField(VSTCharField):
    """
    Field extends :class:`.VSTCharField` and saves file's content as string.

    Value must be text (not binary) and saves in model as is.

    .. note::
        Take effect only in GUI. In API it would behave as :class:`.VSTCharField`.
    """

    __slots__ = ()


class SecretFileInString(FileInStringField):
    """
    Field extends :class:`.FileInStringField`, but hides it's value in admin interface.

    Value must be text (not binary) and saves in model as is.

    .. note::
        Take effect only in GUI. In API it would behave as :class:`.VSTCharField`.
    """

    __slots__ = ()

    def __init__(self, **kwargs):
        kwargs['style'] = {'input_type': 'password'}
        super().__init__(**kwargs)


class BinFileInStringField(FileInStringField):
    """
    Field extends :class:`.FileInStringField`, but work with binary(base64) files.

    .. note::
        Effective only in GUI. Works similar to :class:`.VSTCharField` in API.
    """

    __slots__ = ()


class AutoCompletionField(VSTCharField):
    """
    Field that provides autocompletion on frontend, using specified list of objects.

    :param autocomplete: Autocompletion reference. You can set list/tuple with
                         values or set OpenApi schema definition name.
                         For definition name GUI will find optimal link and
                         will show values based on ``autocomplete_property`` and
                         ``autocomplete_represent`` arguments.
    :type autocomplete: list,tuple,str
    :param autocomplete_property: this argument indicates which attribute will be
                                  get from OpenApi schema definition model as value.
    :type autocomplete_property: str
    :param autocomplete_represent: this argument indicates which attribute will be
                                   get from OpenApi schema definition model as represent value.
    :param use_prefetch: prefetch values on frontend at list-view. Default is ``True``.
    :type use_prefetch: bool

    .. note::
        Effective only in GUI. Works similar to :class:`.VSTCharField` in API.
    """
    __slots__ = 'autocomplete', 'autocomplete_property', 'autocomplete_represent', 'use_prefetch'

    autocomplete: _t.Text
    autocomplete_property: _t.Text
    autocomplete_represent: _t.Text
    use_prefetch: bool

    def __init__(self, **kwargs):
        self.autocomplete = kwargs.pop('autocomplete')
        self.autocomplete_property = None  # type: ignore
        if not isinstance(self.autocomplete, (list, tuple)):
            self.autocomplete_property = kwargs.pop('autocomplete_property', 'id')
            self.autocomplete_represent = kwargs.pop('autocomplete_represent', 'name')
            self.use_prefetch = kwargs.pop('use_prefetch', True)
        super().__init__(**kwargs)


class CommaMultiSelect(VSTCharField):
    """
    Field containing a list of values with specified separator(default: ",").
    Gets list of values from another model or custom list. Provides autocompletion as :class:`.AutoCompletionField`,
    but with comma-lists.
    Suited for property-fields in model where main logic is already implemented or
    with :class:`CharField`.

    :param select: OpenApi schema definition name or list with values.
    :type select: str,tuple,list
    :param select_separator: separator of values. Default is comma.
    :type select_separator: str
    :param select_property,select_represent: work as ``autocomplete_property`` and ``autocomplete_represent``.
                                             Default is ``name``.
    :param use_prefetch: prefetch values on frontend at list-view. Default is ``False``.
    :param make_link: Show value as link to model. Default is ``True``.
    :param dependence: Dictionary, where keys are name of field from the same model, and values are name of query filter
                       .If at least one of the fields that we depend on is non nullable, required and set to null,
                       autocompletion list will be empty and field will be disabled.
    :type dependence: dict


    .. note::
        Effective only in GUI. Works similar to :class:`.VSTCharField` in API.
    """

    __slots__ = (
        'select_model',
        'select_separator',
        'select_property',
        'select_represent',
        'use_prefetch',
        'make_link',
        'dependence'
    )

    select_model: _t.Text
    select_separator: _t.Text
    select_property: _t.Text
    select_represent: _t.Text
    use_prefetch: bool
    make_link: bool
    dependence: DependenceType

    def __init__(self, **kwargs):
        self.select_model = kwargs.pop('select')
        self.select_separator = kwargs.pop('select_separator', ',')
        self.select_property = None  # type: ignore
        if not isinstance(self.select_model, (list, tuple)):
            self.select_property = kwargs.pop('select_property', 'name')
            self.select_represent = kwargs.pop('select_represent', 'name')
        self.use_prefetch = kwargs.pop('use_prefetch', False)
        self.make_link = kwargs.pop('make_link', True)
        self.dependence = kwargs.pop('dependence', None)
        super().__init__(**kwargs)

    def to_internal_value(self, data: _t.Union[_t.Text, _t.Sequence]) -> _t.Text:
        return self.to_representation(data)  # nocv

    def to_representation(self, data: _t.Union[_t.Text, _t.Sequence, _t.Iterator]) -> _t.Text:
        if isinstance(data, str):
            data = map(str, filter(bool, data.split(self.select_separator)))
        return self.select_separator.join(data)


class DynamicJsonTypeField(VSTCharField):
    """
    Field which type is based on another field. It converts value to internal string
    and represent field as json object.

    :param field: field in model which value change will change type of current value.
    :type field: str
    :param types: key-value mapping where key is value of subscribed field and
                  value is type (in OpenApi format) of current field.
    :type type: dict
    :param choices: variants of choices for different subscribed field values.
                    Uses mapping where key is value of subscribed field and
                    value is list with values to choice.
    :type choices: dict


    .. note::
        Effective only in GUI. In API works similar to :class:`.VSTCharField` without value modifications.
    """
    __slots__ = 'field', 'choices', 'types'

    field: _t.Text
    choices: _t.Dict
    types: _t.Dict

    to_json = True

    def __init__(self, **kwargs):
        self.field = kwargs.pop('field')
        self.choices = kwargs.pop('choices', {})
        self.types = kwargs.pop('types', {})
        super().__init__(**kwargs)

    def to_internal_value(self, data):
        return super().to_internal_value(data) if self.to_json else data

    def to_representation(self, value):
        with raise_context():
            value = json.loads(value) if self.to_json else value
        return value


class DependEnumField(DynamicJsonTypeField):
    """
    Field extends :class:`DynamicJsonTypeField` but its value is not transformed to json and would be given as is.
    Useful for :class:`property` in models or for actions.

    :param field: field in model which value change will change type of current value.
    :type field: str
    :param types: key-value mapping where key is value of subscribed field and
                  value is type (in OpenApi format) of current field.
    :type type: dict
    :param choices: variants of choices for different subscribed field values.
                    Uses mapping where key is value of subscribed field and
                    value is list with values to choice.
    :type choices: dict


    .. note::
        Effective only in GUI. In API works similar to :class:`.VSTCharField` without value modification.
    """
    __slots__ = ()
    to_json = False


class DependFromFkField(DynamicJsonTypeField):
    """
    Field extends :class:`DynamicJsonTypeField`. Validates field data by :attr:`.field_attribute`
    chosen in related model. By default, any value of :attr:`.field_attribute` validates as :class:`.VSTCharField`.
    To override this behavior set dict attribute ``{field_attribute value}_fields_mapping`` in related model
     where:

    - **key** - string representation of value type which is received from related instance :attr:`.field_attribute`.
    - **value** - :class:`rest_framework.Field` instance for validation.

    :param field: field in model which value change changes type of current value.
                  Field must be :class:`.FkModelField`.
    :type field: str
    :param field_attribute: attribute of model related model instance with name of type.
    :type field_attribute: str

    .. warning::
        ``field_attribute`` in related model must be :class:`rest_framework.ChoicesField` or
        GUI will show field as simple text.

    """
    __slots__ = ('field', 'field_attribute')

    default_related_field = VSTCharField(allow_null=True, allow_blank=True, default='')

    def __init__(self, **kwargs):
        self.field = kwargs.pop('field')
        self.field_attribute = kwargs.pop('field_attribute')
        super(DynamicJsonTypeField, self).__init__(**kwargs)  # pylint: disable=bad-super-call

    def get_value(self, dictionary: _t.Any) -> _t.Any:
        value = super().get_value(dictionary)
        if value == empty:
            return value  # nocv

        parent_field = self.parent.fields[self.field]  # type: ignore
        related_object: models.Model = parent_field.get_value(dictionary)  # type: ignore
        if not isinstance(related_object, models.Model) and isinstance(parent_field, FkModelField):  # nocv
            # TODO: write test to it
            model_class = get_if_lazy(parent_field.model_class)
            related_object = model_class(**{parent_field.autocomplete_property: related_object})
        related_type = getattr(related_object, self.field_attribute)
        related_field: Field = getattr(
            related_object,
            f'{self.field_attribute}_fields_mapping',
            {related_type: self.default_related_field}
        ).get(related_type, self.default_related_field)
        related_field: Field = copy.deepcopy(related_field)
        related_field.field_name: _t.Text = self.field_name  # type: ignore

        errors: _t.Dict = {}
        primitive_value = related_field.get_value(dictionary)
        try:
            value = related_field.run_validation(primitive_value)
        except ValidationError as exc:
            errors[related_field.field_name] = exc.detail  # type: ignore
        except DjangoValidationError as exc:  # nocv
            errors[related_field.field_name] = get_error_detail(exc)  # type: ignore
        except SkipField:  # nocv
            pass

        if errors:
            raise ValidationError(errors)

        return value


class TextareaField(VSTCharField):
    """
    Field containing multiline string.

    .. note::
        Effective only in GUI. Works similar to :class:`.VSTCharField` in API.
    """

    __slots__ = ()


class HtmlField(VSTCharField):
    """
    Field contains html text and marked as format:html. The field does not validate whether its content is HTML.

    .. warning::
        To avoid vulnerability, do not allow users to modify this data because users ate able to execute their scripts.

    .. note::
        Effective only in GUI. Works similar to :class:`.VSTCharField` in API.
    """

    __slots__ = ()


class FkField(IntegerField):
    """
    Implementation of ForeignKeyField.You can specify which field of a related model will be
    stored in field(default: "id"), and which will represent field on frontend.

    :param select: OpenApi schema definition name.
    :type select: str
    :param autocomplete_property: this argument indicates which attribute will be
                                  get from OpenApi schema definition model as value.
                                  Default is ``id``.
    :type autocomplete_property: str
    :param autocomplete_represent: this argument indicates which attribute will be
                                   get from OpenApi schema definition model as represent value.
                                   Default is ``name``.
    :param use_prefetch: prefetch values on frontend at list-view. Default is ``True``.
    :type use_prefetch: bool
    :param make_link: Show value as link to model. Default is ``True``.
    :type make_link: bool
    :param dependence: Dictionary, where keys are names of a field from the same model,
                       and keys are name of query filter.
                       If at least one of the fields that we depend on is non nullable, required and set to null,
                       autocompletion list will be empty and field will be disabled.
    :type dependence: dict
    :param filters: Dictionary, where keys are names of a field from a related (by this FkField) model,
                    and values are values of that field.
    :type filters: dict

    .. note::
        Intersection of `dependence.values()` and `filters.keys()` will throw error to prevent ambiguous filtering.
    .. note::
        Effective only in GUI. Works similar to :class:`rest_framework.IntegerField` in API.
    """
    __slots__ = (
        'select_model',
        'autocomplete_property',
        'autocomplete_represent',
        'use_prefetch',
        'make_link',
        'dependence'
    )

    select_model: _t.Text
    autocomplete_property: _t.Text
    autocomplete_represent: _t.Text
    use_prefetch: bool
    make_link: bool
    dependence: DependenceType
    default_error_messages = {
        'ambiguous_filter': 'Ambiguous filtering, use different filters for dependencies and filters',
    }

    def __init__(self, **kwargs):
        self.select_model = kwargs.pop('select')
        self.autocomplete_property = kwargs.pop('autocomplete_property', 'id')
        self.autocomplete_represent = kwargs.pop('autocomplete_represent', 'name')
        self.use_prefetch = kwargs.pop('use_prefetch', True)
        self.make_link = kwargs.pop('make_link', True)
        self.dependence = kwargs.pop('dependence', None)
        self.filters = kwargs.pop('filters', None)
        super().__init__(**kwargs)
        if self.filters and self.dependence and set(self.dependence.values()) & set(self.filters.keys()):
            self.fail('ambiguous_filter')


class FkModelField(FkField):
    """
    Extends :class:`.FkField`, but stores referred model class.
    This field is useful for :class:`django.db.models.ForeignKey` fields in model to set.

    :param select: model class (based on :class:`vstutils.models.BModel`) or serializer class
                   which used in API and has path in OpenApi schema.
    :type select: vstutils.models.BModel,vstutils.api.serializers.VSTSerializer
    :param autocomplete_property: this argument indicates which attribute will be
                                  get from OpenApi schema definition model as value.
                                  Default is ``id``.
    :type autocomplete_property: str
    :param autocomplete_represent: this argument indicates which attribute will be
                                   get from OpenApi schema definition model as represent value.
                                   Default is ``name``.
    :param use_prefetch: prefetch values on frontend at list-view. Default is ``True``.
    :param make_link: Show value as link to model. Default is ``True``.


    .. warning::
        Model class get object from database during `.to_internal_value` execution. Be careful on mass save executions.

    .. warning::
        Permissions to model which is referred by this field, are not to be checked.
        You should check it manually in signals or validators.

    """

    __slots__ = ('model_class',)

    model_class: _t.Type[models.Model]

    def __init__(self, **kwargs):
        select = kwargs.pop('select')
        if '__extra_metadata__' in dir(select):
            self.model_class = select
            kwargs['select'] = self._get_lazy_select_name_from_model()
        elif isinstance(select, str):
            select = select.split('.')
            assert len(select) == 2, "'select' must match 'app_name.model_name' pattern."
            self.model_class = SimpleLazyObject(lambda: apps.get_model(require_ready=True, *select))
            kwargs['select'] = self._get_lazy_select_name_from_model()
        elif issubclass(select, ModelSerializer):
            self.model_class = select.Meta.model
            kwargs['select'] = select.__name__.replace('Serializer', '')
        else:  # nocv
            raise Exception(
                'Argument "select" must be '
                'rest_framework.serializers.ModelSerializer or '
                'vstutils.models.BModel subclass or '
                'string matched "app_name.model_name" pattern.'
            )
        super().__init__(**kwargs)
        # Remove min/max validators from integer field.
        self.validators = self.validators[:-2]

    def _get_lazy_select_name_from_model(self):
        # pylint: disable=invalid-name
        return SimpleLazyObject(
            lambda: self.model_class.get_list_serializer_name().split('Serializer')[0]
        )

    @functools.lru_cache()
    def _get_data_from_model(self, value):
        self.model_class = get_if_lazy(self.model_class)
        return self.model_class.objects.get(**{self.autocomplete_property: value})

    def get_value(self, dictionary: _t.Any) -> _t.Any:
        value = super().get_value(dictionary)
        if value is not empty and value is not None:
            value = self._get_data_from_model(value)
        return value

    def to_internal_value(self, data: _t.Union[models.Model, int]) -> _t.Union[models.Model]:  # type: ignore[override]
        if isinstance(data, self.model_class):
            return data
        elif isinstance(data, (int, str)):  # nocv
            # deprecated
            return self._get_data_from_model(data)
        self.fail('Unknown datatype')  # nocv

    def to_representation(self, value: _t.Union[int, models.Model]) -> _t.Any:
        self.model_class = get_if_lazy(self.model_class)
        if isinstance(value, self.model_class):
            return getattr(value, self.autocomplete_property)
        else:
            return value  # type: ignore


class UptimeField(IntegerField):
    """
    Field for some uptime(time duration), in seconds, for example.

    .. note::
        Effective only in GUI. Works similar to :class:`rest_framework.IntegerField` in API.

    """

    __slots__ = ()


class RedirectFieldMixin:
    """
    Field mixin indicates that this field is used to send redirect address to frontend after some action.

    :param operation_name: prefix for operation_id, for example if operation_id is `history_get`
           then operation_name is `history`
    :type operation_name: str
    :param depend_field: name of the field that we depend on, its' value will be used for operation_id
    :type depend_field: str
    :param concat_field_name: if True then name of the field will be added at the end of operation_id
    :type concat_field_name: bool
    """
    redirect: bool = True

    __slots__ = ('operation_name', 'depend_field', 'concat_field_name')

    def __init__(self, **kwargs):
        self.operation_name = kwargs.pop('operation_name', None)
        self.depend_field = kwargs.pop('depend_field', None)
        self.concat_field_name = kwargs.pop('concat_field_name', False)
        super().__init__(**kwargs)


class RedirectIntegerField(RedirectFieldMixin, IntegerField):
    """
    Field for redirect by id. Often used in actions for redirect after execution.

    .. note::
        Effective only in GUI. Works similar to :class:`rest_framework.IntegerField` in API.

    """

    __slots__ = ()


class RedirectCharField(RedirectFieldMixin, CharField):
    """
    Field for redirect by string. Often used in actions for redirect after execution.

    .. note::
        Effective only in GUI. Works similar to :class:`rest_framework.IntegerField` in API.

    """

    __slots__ = ()


class NamedBinaryFileInJsonField(VSTCharField):
    """
    Field that takes JSON with properties:
    * name - string - name of file;
    * mediaType - string - MIME type of file
    * content - base64 string - content of file.

    This field is useful for saving binary files with their names in :class:`django.db.models.CharField`
    or :class:`django.db.models.TextField` model fields. All manipulations with decoding and encoding
    binary content data executes on client. This imposes reasonable limits on file size.

    Additionally, this field can construct :class:`django.core.files.uploadedfile.SimpleUploadedFile`
    from incoming JSON and store it as file in :class:`django.db.models.FileField` if `file` argument is set to `True`

    Attrs:
    :attr:`NamedBinaryInJsonField.file`: if True, accept only subclasses of File as input. If False, accept only string
    input. Default: False.

    .. note::
        Effective only in GUI. Works similar to :class:`.VSTCharField` in API.

    """

    __slots__ = ('file',)

    __valid_keys = ('name', 'content', 'mediaType')
    default_error_messages = {
        'not a JSON': 'value is not a valid JSON',
        'missing key': 'key {missing_key} is missing',
        'invalid key': 'invalid key {invalid_key}',
    }
    file_field = FieldFile

    def __init__(self, **kwargs):
        self.file = kwargs.pop('file', False)
        super(NamedBinaryFileInJsonField, self).__init__(**kwargs)

    def validate_value(self, data: _t.Dict):
        if not isinstance(data, dict):
            self.fail('not a JSON')
        invalid_keys = [
            k
            for k in data.keys()
            if k not in self.__valid_keys
        ]

        if invalid_keys:
            self.fail('invalid key', invalid_key=invalid_keys[0])

        for key in self.__valid_keys:
            if key not in data:
                self.fail('missing key', missing_key=key)

    @raise_context_decorator_with_default(default=DEFAULT_NAMED_FILE_DATA)
    def to_representation(self, value) -> _t.Dict[_t.Text, _t.Optional[_t.Any]]:
        if self.file:
            return {'content': value.url, 'name': value.name, 'mediaType': ''}
        else:
            result = json.loads(value)
            if not result.get('mediaType'):
                result['mediaType'] = None
            return result

    def run_validation(self, data=empty):
        (is_empty_value, data) = self.validate_empty_values(data)
        if is_empty_value:
            return data  # nocv
        self.validate_value(data)
        self.run_validators(data)
        return self.to_internal_value(data)

    def _handle_file(self, file):
        if not file['mediaType'] \
           and (file['content'].startswith('/') or file['content'].startswith('http')) \
           and file['content'].endswith(file['name']) \
           and self.root.instance:
            return self.file_field(
                self.root.instance,
                self.root.instance._meta.get_field(self.field_name),
                file['name']
            )
        return SimpleUploadedFile(
            name=file['name'],
            content=base64.b64decode(file['content']),
            content_type=file['mediaType']
        )

    def to_internal_value(self, data) -> _t.Any:
        if self.file and data.get('content', None):
            return self._handle_file(data)
        return super().to_internal_value(data)


class NamedBinaryImageInJsonField(NamedBinaryFileInJsonField):
    """
    Extends :class:`.NamedBinaryFileInJsonField` to represent image on frontend
    (if binary image is valid).Validate this field with
    :class:`vstutils.api.validators.ImageValidator`.
    """

    __slots__ = ()


class MultipleNamedBinaryFileInJsonField(NamedBinaryFileInJsonField):
    """
    Extends :class:`.NamedBinaryFileInJsonField` but uses list of JSONs.
    Allows to operate with multiple files as list of :class:`NamedBinaryFileInJsonField`.

    Attrs:
    :attr:`NamedBinaryInJsonField.file`: if True, accept only subclasses of File as input. If False, accept only string
    input. Default: False.
    """

    __slots__ = ()
    default_error_messages = {
        'not a list': 'value is not a valid list',
    }
    file_field = vst_model_fields.MultipleFieldFile

    def run_validators(self, value: _t.Any) -> None:
        if isinstance(value, (list, tuple)):
            for one_value in value:
                super().run_validators(one_value)
        else:  # nocv
            super().run_validators(value)

    def to_internal_value(self, data: _t.List) -> _t.Union[_t.Text, list]:  # type: ignore
        if data is not None:
            if not isinstance(data, list):
                self.fail('not a list')
            for file in data:
                self.validate_value(file)
            if self.file:
                return [
                    self._handle_file(file)
                    for file in data
                ]
        return VSTCharField.to_internal_value(self, data)

    def run_validation(self, data=empty):
        (is_empty_value, data) = self.validate_empty_values(data)
        if is_empty_value:
            return data  # nocv
        self.run_validators(data)
        return self.to_internal_value(data)

    @raise_context_decorator_with_default(default=[])
    def to_representation(self, value) -> _t.List[_t.Dict[_t.Text, _t.Any]]:  # type: ignore
        if self.file:
            return [
                {'content': file.url, 'name': file.name, 'mediaType': ''}
                for file in value
            ]
        return json.loads(value)


class MultipleNamedBinaryImageInJsonField(MultipleNamedBinaryFileInJsonField):
    """
    Extends :class:`.MultipleNamedBinaryFileInJsonField` but uses list of JSONs.
    Used for operating with multiple images and works as list of :class:`NamedBinaryImageInJsonField`.
    """

    __slots__ = ()


class PasswordField(CharField):
    """
    Extends `CharField <https://www.django-rest-framework.org/api-guide/fields/#charfield>`_
    but in schema set format to `password`.
    Show all characters as asterisks instead of real value in GUI.
    """

    def __init__(self, *args, **kwargs):
        kwargs['write_only'] = True
        kwargs['style'] = kwargs.get('style', {})
        kwargs['style']['input_type'] = 'password'
        super(PasswordField, self).__init__(*args, **kwargs)


def _handle_related_value_decorator(func):
    def wrapper(self, value, default, *args, **kwargs):
        if not value:
            return default
        try:
            return func(self, value, default, *args, **kwargs)
        except Exception as err:
            logging.error(err)
            return default
    return wrapper


class RelatedListField(VSTCharField):
    """
    Extends class 'vstutils.api.fields.VSTCharField'. With this field you can output reverse ForeignKey relation
    as a list of related instances.
    To use it, you specify 'related_name' kwarg(related_manager for reverse ForeignKey)
    and 'fields' kwarg(list or tuple of fields from related model, which needs to be included)

    :param related_name: name of a related manager for reverse foreign key
    :type related_name: str
    :param fields: list of related model fields.
    :type fields: list[str], tuple[str]
    :param view_type: determines how field are represented on frontend. Must be either 'list' or 'table'.
    :type view_type: str
    :param fields_custom_handlers_mapping: includes custom handlers, where key: field_name, value: callable_obj that
                                           takes params: instance[dict], fields_mapping[dict], model, field_name[str]
    :type fields_custom_handlers_mapping: dict
    """

    def __init__(self, related_name: _t.Text, fields: _t.Union[_t.Tuple, _t.List], view_type: str = 'list', **kwargs):
        kwargs['read_only'] = True
        kwargs['source'] = "*"
        self.fields_custom_handlers_mapping = kwargs.pop('fields_custom_handlers', {})
        super().__init__(**kwargs)
        # fields for 'values' in qs
        assert isinstance(fields, (tuple, list)), "fields must be list or tuple"
        assert fields, "fields must have one or more values"
        assert view_type in ('list', 'table')
        self.fields = fields
        self.related_name = related_name
        self.view_type = view_type

    @_handle_related_value_decorator
    def _handle_named_bin_text(self, value, default):
        return orjson.loads(value)

    @_handle_related_value_decorator
    def _handle_file_or_image_field(self, value, default, field):
        return {
            "content": field.storage.url(value),
            "name": value,
            'mediaType': '',
        }

    @_handle_related_value_decorator
    def _handle_files_or_images_field(self, list_values, default, field):
        return [
            self._handle_file_or_image_field(value, DEFAULT_NAMED_FILE_DATA, field)
            for value in list_values
        ]

    def handle_values_item_iteration(self, instance, fields_mapping, field_name, value, full_field_name=None):
        full_field_name = full_field_name or field_name
        field = fields_mapping.get(field_name, None)

        if '__' in field_name:
            parent_field_name, child_field_name = field_name.split('__', maxsplit=1)
            field = fields_mapping.get(parent_field_name, None)

            if isinstance(field, models.ForeignKey):
                self.handle_values_item_iteration(
                    instance,
                    self.get_model_fields_mapping(field.related_model),
                    child_field_name,
                    value,
                    full_field_name
                )
        elif isinstance(field, vst_model_fields.NamedBinaryFileInJSONField):
            instance[full_field_name] = self._handle_named_bin_text(value, DEFAULT_NAMED_FILE_DATA)

        elif isinstance(field, vst_model_fields.MultipleNamedBinaryFileInJSONField):
            instance[full_field_name] = self._handle_named_bin_text(value, [])

        elif isinstance(field, vst_model_fields.MultipleFileMixin):
            instance[full_field_name] = self._handle_files_or_images_field(value, [], fields_mapping[field_name])

        elif isinstance(field, vst_model_fields.FileField):
            instance[full_field_name] = self._handle_file_or_image_field(
                value,
                DEFAULT_NAMED_FILE_DATA,
                fields_mapping[field_name],
            )

    def handle_values_item(self, instance, fields_mapping, model):
        for field_name, value in instance.items():
            if field_name in self.fields_custom_handlers_mapping:
                self.fields_custom_handlers_mapping[field_name](self, instance, fields_mapping, model, field_name)
            else:
                self.handle_values_item_iteration(instance, fields_mapping, field_name, value)

        return instance

    def get_model_fields_mapping(self, model):
        return {f.name: f for f in model._meta.fields}

    def to_representation(self, value: _t.Type[models.Model]) -> _t.Tuple[_t.Dict]:  # type: ignore[override]
        queryset = getattr(value, self.related_name).all()

        handler = functools.partial(
            self.handle_values_item,
            fields_mapping=self.get_model_fields_mapping(queryset.model),
            model=value,
        )
        # get related mapping with id and name of instances
        return lazy(lambda: tuple(map(handler, getattr(value, self.related_name).values(*self.fields))), tuple)()


class RatingField(FloatField):
    """
    Extends class 'rest_framework.serializers.FloatField'. This field represents a rating form input on frontend.
    Grading limits can be specified with 'min_value=' and 'max_value=', defaults are 0 to 5.Minimal step between
    grades are specified in 'step=', default - 1.Frontend visual representation can be chosen
    with 'front_style=', available variants are listed in 'self.valid_front_styles'.

    for 'slider' front style, you can specify slider color, by passing valid color to 'color='.
    for 'fa_icon' front style, you can specify FontAwesome icon that would be used for displaying rating, by passing a
    valid FontAwesome icon code to 'fa_class='.

    :param min_value: minimal level
    :type min_value: float, int
    :param max_value: maximal level
    :type max_value: float, int
    :param step: minimal step between levels
    :type step: float, int
    :param front_style: visualization on frontend field. Allowed: ['stars', 'slider', 'fa_icon'].
    :type front_style: str
    :param color: color of rating element (star, icon or slider) in css format
    :type color: str
    :param fa_class: FontAwesome icon code
    :type fa_class: str
    """
    valid_front_styles = (
        'stars',
        'slider',
        'fa_icon',
    )

    def __init__(
            self,
            min_value: float = 0,
            max_value: float = 5,
            step: float = 1,
            front_style: _t.Text = 'stars',
            **kwargs,
    ):
        assert front_style in self.valid_front_styles, f"front_style should be one of {self.valid_front_styles}"
        self.front_style = front_style
        self.color = kwargs.pop('color', None)
        assert isinstance(self.color, str) or self.color is None, "color should be str"
        self.fa_class = kwargs.pop('fa_class', None)
        assert isinstance(self.fa_class, str) or self.fa_class is None, "fa_class should be str"
        self.step = step
        assert not (step != 1 and front_style != 'slider'), 'custom step can be used only with front_style "slider"'
        super(RatingField, self).__init__(min_value=min_value, max_value=max_value, **kwargs)



from django.apps import AppConfig


class VSTUtilsAppConfig(AppConfig):
    name = 'vstutils.api'
    label = 'vstutils_api'
    verbose_name = "vstutils"

    def ready(self):
        super().ready()
        self.module.autodiscover()



# pylint: disable=import-error,signature-differs
from django_celery_beat.schedulers import DatabaseScheduler

from .utils import Lock


class SingletonDatabaseScheduler(DatabaseScheduler):
    scheduler_lock = None

    def tick(self, *args, **kwargs):
        if self.scheduler_lock is None:
            try:
                self.scheduler_lock = Lock(Lock.SCHEDULER, timeout=120.0)
            except Lock.AcquireLockException:
                return 60.0
        self.scheduler_lock.prolong()
        return super().tick(*args, **kwargs)

    def close(self):
        if self.scheduler_lock is not None:
            self.scheduler_lock.release()

        return super().close()



# pylint: disable=abstract-method,unused-import,undefined-variable

import os
import logging
from collections import OrderedDict

from django.conf import settings
from django.core.management.base import (
    BaseCommand as _BaseCommand,
    CommandError as CmdError,
)


logger = logging.getLogger(settings.VST_PROJECT)
logger_lib = logging.getLogger(settings.VST_PROJECT_LIB)


class BaseCommand(_BaseCommand):
    interactive = False
    requires_system_checks = False
    keep_base_opts = False
    help = "Service command for web-application"

    class CommandError(CmdError):
        pass

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.fromfile_prefix_chars = '@'
        parser.add_argument(
            '-l', '--log-level',
            action='store',
            dest='log-level',
            default=False,
            type=str,
            help='Set logs level [debug|warning|error|critical]')
        if self.interactive:
            parser.add_argument(
                '--noinput', '--no-input',
                action='store_false', dest='interactive', default=True,
                help="Do NOT prompt the user for input of any kind.",
            )

    def _print(self, msg, style=None):
        style = style or 'HTTP_INFO'
        style = getattr(self.style, style, str)
        self.stdout.write(style(msg))

    def _settings(self, value, default=None):
        return getattr(settings, value, default)

    def _get_versions(self):
        versions = OrderedDict(django=super().get_version())
        versions[self._settings('VST_PROJECT')] = self._settings('PROJECT_VERSION')
        if self._settings('VST_PROJECT') != self._settings('VST_PROJECT_LIB'):
            versions[self._settings('VST_PROJECT_LIB')] = self._settings(
                'PROJECT_LIB_VERSION'
            )
        if self._settings('VST_PROJECT') != "vstutils":
            versions['vstutils'] = self._settings('VSTUTILS_VERSION')
        return versions

    def get_version(self):
        return ' '.join([
            f'{k.title()}={v}' for k, v in self._get_versions().items()
        ])

    def ask_user(self, message, default=None):
        # pylint: disable=import-error
        if getattr(self, 'interactive_mode', False):
            return input(message) or default
        return default

    def ask_user_bool(self, message, default=True):
        reply = self.ask_user(message, 'yes' if default else 'no').lower()
        if reply in ['y', 'yes']:
            return True
        elif reply in ['n', 'no']:
            return False

    def handle(self, *args, **options):
        # pylint: disable=invalid-name
        self.interactive_mode = options.pop('interactive', False)
        LOG_LEVEL = settings.LOG_LEVEL
        if options.get('log-level', False):
            LOG_LEVEL = options.pop('log-level', LOG_LEVEL)
        logger.setLevel(LOG_LEVEL.upper())
        logger_lib.setLevel(LOG_LEVEL.upper())
        self.LOG_LEVEL = LOG_LEVEL.upper()
        os.environ.setdefault('DJANGO_LOG_LEVEL', self.LOG_LEVEL)



from typing import Dict, Type, Text, Any
from collections import OrderedDict

from django.http import FileResponse
from drf_yasg.inspectors.base import FieldInspector, NotHandled
from drf_yasg.inspectors.field import ReferencingSerializerInspector
from drf_yasg import openapi
from drf_yasg.inspectors.query import CoreAPICompatInspector
from rest_framework.fields import Field, JSONField

from .. import fields, serializers, validators
from ...models.base import get_first_match_name


# Extra types

# Extra formats
FORMAT_FILE = openapi.TYPE_FILE
FORMAT_SECRET_FILE = 'secretfile'  # nosec
FORMAT_BIN_FILE = 'binfile'
FORMAT_NAMED_BIN_FILE = 'namedbinfile'
FORMAT_NAMED_BIN_IMAGE = 'namedbinimage'
FORMAT_MULTIPLE_NAMED_BIN_FILE = 'multiplenamedbinfile'
FORMAT_MULTIPLE_NAMED_BIN_IMAGE = 'multiplenamedbinimage'
FORMAT_AUTOCOMPLETE = 'autocomplete'
FORMAT_FK_AUTOCOMPLETE = 'fk_autocomplete'
FORMAT_MULTISELECT = 'multiselect'
FORMAT_HTML = 'html'
FORMAT_JSON = 'json'
FORMAT_TEXTAREA = 'textarea'
FORMAT_DYN = 'dynamic'
FORMAT_DYN_FK = 'dynamic_fk'
FORMAT_FK = 'fk'
FORMAT_UPTIME = 'uptime'
FORMAT_RELATED_LIST = 'related_list'
FORMAT_RATING = 'rating'


# Base types
basic_type_info: Dict[Type[Field], Dict[Text, Any]] = OrderedDict()
basic_type_info[fields.FileInStringField] = {
    'type': openapi.TYPE_STRING,
    'format': FORMAT_FILE
}
basic_type_info[fields.SecretFileInString] = {
    'type': openapi.TYPE_STRING,
    'format': FORMAT_SECRET_FILE
}
basic_type_info[fields.BinFileInStringField] = {
    'type': openapi.TYPE_STRING,
    'format': FORMAT_BIN_FILE
}
basic_type_info[fields.NamedBinaryFileInJsonField] = {
    'type': openapi.TYPE_STRING,
    'format': FORMAT_NAMED_BIN_FILE
}
basic_type_info[fields.MultipleNamedBinaryFileInJsonField] = {
    'type': openapi.TYPE_STRING,
    'format': FORMAT_MULTIPLE_NAMED_BIN_FILE
}
basic_type_info[fields.HtmlField] = {
    'type': openapi.TYPE_STRING,
    'format': FORMAT_HTML
}
basic_type_info[serializers.JsonObjectSerializer] = {
    'type': openapi.TYPE_OBJECT,
    'format': FORMAT_JSON
}
basic_type_info[JSONField] = {
    'type': openapi.TYPE_STRING,
    'format': FORMAT_JSON
}
basic_type_info[fields.TextareaField] = {
    'type': openapi.TYPE_STRING,
    'format': FORMAT_TEXTAREA
}
basic_type_info[fields.UptimeField] = {
    'type': openapi.TYPE_INTEGER,
    'format': FORMAT_UPTIME
}
basic_type_info[fields.RedirectIntegerField] = {
    'type': openapi.TYPE_INTEGER
}
basic_type_info[fields.RedirectCharField] = {
    'type': openapi.TYPE_STRING
}
basic_type_info[fields.PasswordField] = {
    'type': openapi.TYPE_STRING,
    'format': openapi.FORMAT_PASSWORD
}


def field_have_redirect(field, **kwargs):
    if not getattr(field, 'redirect', False):
        return kwargs

    if kwargs.get('additionalProperties', None) is None:
        kwargs['additionalProperties'] = {}
    kwargs['additionalProperties']['redirect'] = {
        'operation_name': getattr(field, 'operation_name', None) or field.field_name,
        'depend_field': getattr(field, 'depend_field', None),
        'concat_field_name': getattr(field, 'concat_field_name', False),
    }

    return kwargs


def field_extra_handler(field, **kwargs):
    kwargs = field_have_redirect(field, **kwargs)
    return kwargs


def _get_handled_props(group_mapping):
    for group_name, props in (group_mapping or {}).items():
        if group_name == '':
            continue
        yield from props


class VSTFieldInspector(FieldInspector):
    def field_to_swagger_object(self, field, swagger_object_type, use_references, **kw):
        # pylint: disable=invalid-name,unused-variable
        type_info = basic_type_info.get(field.__class__, None)
        if type_info is None:
            return NotHandled

        SwaggerType, ChildSwaggerType = self._get_partial_types(
            field, swagger_object_type, use_references, **kw
        )
        return SwaggerType(**field_extra_handler(field, **type_info))


class AutoCompletionFieldInspector(FieldInspector):
    def field_to_swagger_object(self, field, swagger_object_type, use_references, **kw):
        # pylint: disable=invalid-name
        if not isinstance(field, fields.AutoCompletionField):
            return NotHandled

        SwaggerType, _ = self._get_partial_types(
            field, swagger_object_type, use_references, **kw
        )
        kwargs = {
            'type': openapi.TYPE_STRING,
            'format': FORMAT_AUTOCOMPLETE
        }
        if isinstance(field.autocomplete, (list, tuple)):
            kwargs['enum'] = list(field.autocomplete)
        else:
            kwargs['format'] = FORMAT_FK_AUTOCOMPLETE
            kwargs['additionalProperties'] = {
                'model': openapi.SchemaRef(
                    self.components.with_scope(openapi.SCHEMA_DEFINITIONS),
                    field.autocomplete, ignore_unresolved=True
                ),
                'value_field': field.autocomplete_property,
                'view_field': field.autocomplete_represent,
                'usePrefetch': field.use_prefetch,
            }

        return SwaggerType(**field_extra_handler(field, **kwargs))


class DynamicJsonTypeFieldInspector(FieldInspector):
    def field_to_swagger_object(self, field, swagger_object_type, use_references, **kw):
        # pylint: disable=unused-variable,invalid-name
        if not isinstance(field, fields.DynamicJsonTypeField):
            return NotHandled

        SwaggerType, ChildSwaggerType = self._get_partial_types(
            field, swagger_object_type, use_references, **kw
        )
        additionalProperties = {'field': field.field}

        if isinstance(field, fields.DependFromFkField):
            field_format = FORMAT_DYN_FK
            additionalProperties['field_attribute'] = field.field_attribute

        else:
            field_format = FORMAT_DYN
            additionalProperties['choices'] = field.choices
            additionalProperties['types'] = field.types

        kwargs = {
            'type': openapi.TYPE_STRING,
            'format': field_format,
            'additionalProperties': additionalProperties
        }

        return SwaggerType(**field_extra_handler(field, **kwargs))


class FkFieldInspector(FieldInspector):
    def field_to_swagger_object(self, field, swagger_object_type, use_references, **kw):
        # pylint: disable=unused-variable,invalid-name
        if not isinstance(field, fields.FkField):
            return NotHandled

        SwaggerType, ChildSwaggerType = self._get_partial_types(
            field, swagger_object_type, use_references, **kw
        )
        kwargs = {
            'type': openapi.TYPE_INTEGER,
            'format': FORMAT_FK,
            'additionalProperties': {
                'model': openapi.SchemaRef(
                    self.components.with_scope(openapi.SCHEMA_DEFINITIONS),
                    field.select_model, ignore_unresolved=True
                ),
                'value_field': field.autocomplete_property,
                'view_field': field.autocomplete_represent,
                'usePrefetch': field.use_prefetch,
                'makeLink': field.make_link,
                'dependence': field.dependence,
                'filters': field.filters
            }
        }

        return SwaggerType(**field_extra_handler(field, **kwargs))


class CommaMultiSelectFieldInspector(FieldInspector):
    def field_to_swagger_object(self, field, swagger_object_type, use_references, **kw):
        # pylint: disable=unused-variable,invalid-name
        if not isinstance(field, fields.CommaMultiSelect):
            return NotHandled

        SwaggerType, ChildSwaggerType = self._get_partial_types(
            field, swagger_object_type, use_references, **kw
        )
        kwargs = {
            'type': openapi.TYPE_STRING,
            'format': FORMAT_MULTISELECT,
            'additionalProperties': {
                'model': openapi.SchemaRef(
                    self.components.with_scope(openapi.SCHEMA_DEFINITIONS),
                    field.select_model, ignore_unresolved=True
                ),
                'value_field': field.select_property,
                'view_field': field.select_represent,
                'view_separator': field.select_separator,
                'usePrefetch': field.use_prefetch,
                'makeLink': field.make_link,
                'dependence': field.dependence,
            }
        }

        return SwaggerType(**field_extra_handler(field, **kwargs))


class RelatedListFieldInspector(FieldInspector):
    def field_to_swagger_object(self, field, swagger_object_type, use_references, **kw):
        # pylint: disable=unused-variable,invalid-name
        if not isinstance(field, fields.RelatedListField):
            return NotHandled

        SwaggerType, ChildSwaggerType = self._get_partial_types(
            field, swagger_object_type, use_references, **kw
        )
        kwargs = {
            'type': openapi.TYPE_STRING,
            'format': FORMAT_RELATED_LIST,
            'additionalProperties': {
                'fields': field.fields,
                'viewType': field.view_type,
            }
        }

        return SwaggerType(**field_extra_handler(field, **kwargs))


class RatingFieldInspector(FieldInspector):
    def field_to_swagger_object(self, field, swagger_object_type, use_references, **kw):
        # pylint: disable=unused-variable,invalid-name
        if not isinstance(field, fields.RatingField):
            return NotHandled

        SwaggerType, ChildSwaggerType = self._get_partial_types(
            field, swagger_object_type, use_references, **kw
        )
        kwargs = {
            'type': openapi.TYPE_NUMBER,
            'format': FORMAT_RATING,
            'additionalProperties': {
                'min_value': field.min_value,
                'max_value': field.max_value,
                'step': field.step,
                'style': field.front_style,
                'color': field.color,
                'fa_class': field.fa_class
            }
        }
        return SwaggerType(**field_extra_handler(field, **kwargs))


class NamedBinaryImageInJsonFieldInspector(FieldInspector):
    def field_to_swagger_object(self, field, swagger_object_type, use_references, **kw):
        # pylint: disable=unused-variable,invalid-name
        if isinstance(field, fields.NamedBinaryImageInJsonField):
            img_format = FORMAT_NAMED_BIN_IMAGE
        elif isinstance(field, fields.MultipleNamedBinaryImageInJsonField):
            img_format = FORMAT_MULTIPLE_NAMED_BIN_IMAGE
        else:
            return NotHandled

        SwaggerType, ChildSwaggerType = self._get_partial_types(
            field, swagger_object_type, use_references, **kw
        )
        kwargs = {
            'type': openapi.TYPE_STRING,
            'format': img_format,
            'additionalProperties': {},
        }
        for validator in field.validators:
            if isinstance(validator, validators.ImageResolutionValidator):
                kwargs['additionalProperties'].update(
                    {
                        'min_width': validator.min_width,
                        'max_width': validator.max_width,
                        'min_height': validator.min_height,
                        'max_height': validator.max_height,
                        'extensions': validator.extensions
                    }
                )

        return SwaggerType(**field_extra_handler(field, **kwargs))


class NestedFilterInspector(CoreAPICompatInspector):
    def get_filter_parameters(self, filter_backend):  # nocv
        subaction_list_actions = [
            f'{name}_list'
            for name in getattr(self.view, '_nested_args', {}).keys()
        ]
        if self.view.action not in subaction_list_actions:
            return NotHandled
        if self.method != 'GET':
            return NotHandled
        nested_view = getattr(self.view, self.view.action, None)
        nested_view_filter_class = getattr(nested_view, '_nested_filter_class', None)
        filter_class = getattr(self.view, 'filterset_class', getattr(self.view, 'filter_class', None))
        self.view.filter_class = nested_view_filter_class
        result = super().get_filter_parameters(filter_backend)
        self.view.filter_class = filter_class
        return result


class VSTReferencingSerializerInspector(ReferencingSerializerInspector):
    def get_serializer_ref_name(self, serializer: Any):
        if isinstance(serializer, serializers.serializers.ListSerializer):
            return super().get_serializer_ref_name(serializer.child)
        return super().get_serializer_ref_name(serializer)

    def handle_schema(self, field: Any):
        ref_name = self.get_serializer_ref_name(field)
        definitions = self.components.with_scope(openapi.SCHEMA_DEFINITIONS)

        if ref_name not in definitions:
            return

        schema = definitions[ref_name]

        if getattr(schema, '_handled', False):
            return

        schema_properties = tuple(schema['properties'].keys())
        serializer_class = schema._NP_serializer  # pylint: disable=protected-access

        non_bulk_methods = getattr(serializer_class, '_non_bulk_methods', None)
        schema_properties_groups = OrderedDict(
            getattr(serializer_class, '_schema_properties_groups', None) or {'': schema_properties}
        )
        view_field_name = getattr(serializer_class, '_view_field_name', None)

        if view_field_name is None and schema_properties:
            view_field_name = get_first_match_name(schema_properties, schema_properties[0])

        if schema_properties_groups:
            not_handled = set(schema['properties']) - set(_get_handled_props(schema_properties_groups))

            if not_handled:
                schema_properties_groups[''] = [
                    prop
                    for prop in schema['properties']
                    if prop in not_handled
                ]

        schema['x-properties-groups'] = schema_properties_groups
        schema['x-view-field-name'] = view_field_name

        if non_bulk_methods:
            schema['x-non-bulk-methods'] = non_bulk_methods

        schema._handled = True  # pylint: disable=protected-access

    def field_to_swagger_object(self, field: Any, swagger_object_type: Any, use_references: Any, **kwargs: Any):
        if isinstance(field, FileResponse):
            return openapi.Schema(type='file')

        result = super().field_to_swagger_object(field, swagger_object_type, use_references, **kwargs)

        if result != NotHandled:
            self.handle_schema(field)

        return result



import typing as _t

from rest_framework.filters import BaseFilterBackend
from django_filters.rest_framework.backends import DjangoFilterBackend as BaseDjangoFilterBackend
from django_filters.rest_framework import filters
from django_filters import compat
from django.db import models
from vstutils.utils import raise_context


class DjangoFilterBackend(BaseDjangoFilterBackend):
    def get_coreschema_field(self, field):
        if isinstance(field, filters.NumberFilter):
            field_cls = compat.coreschema.Number
        elif isinstance(field, filters.BooleanFilter):
            field_cls = compat.coreschema.Boolean
        else:
            field_cls = compat.coreschema.String
        return field_cls(
            description=str(field.extra.get('help_text', ''))
        )


# Call standart filtering
class VSTFilterBackend(BaseFilterBackend):
    """
    A base filter backend class to be inherited from.
    Example:

        .. sourcecode:: python

            from django.utils import timezone
            from django.db.models import Value, DateTimeField

            from vstutils.api.filter_backends import VSTFilterBackend

            CurrentTimeFilterBackend(VSTFilterBackend):
                def filter_queryset(self, request, queryset, view):
                    return queryset.annotate(current_time=Value(timezone.now(), output_field=DateTimeField())

        In this example Filter Backend annotates time in current timezone to any connected
        model's queryset.
    """
    __slots__ = ()
    required = False

    def filter_queryset(self, request, queryset, view):
        raise NotImplementedError  # nocv

    def get_schema_fields(self, view):
        """
        You can also make the filter controls available to the schema autogeneration that REST framework provides,
        by implementing this method. The method should return a list of coreapi.Field instances.
        """
        # pylint: disable=unused-argument
        return []


class HideHiddenFilterBackend(VSTFilterBackend):
    """Filter Backend that hides all objects with hidden=True from the queryset"""
    __slots__ = ()
    required = True

    def filter_queryset(self, request, queryset, view):
        # pylint: disable=unused-argument
        """
        Clear objects with hidden attr from queryset.
        """
        return getattr(queryset, 'cleared', queryset.all)()


class SelectRelatedFilterBackend(VSTFilterBackend):
    """
    Filter Backend that will automatically call prefetch_related and select_related on all relations in queryset.
    """
    __slots__ = ()
    required = True
    fields_fetch_map = {
        'select': (models.ForeignKey, models.OneToOneField),
        'prefetch': (models.ManyToManyField, models.ManyToManyField.rel_class)
    }

    def filter_model_fields(self, view, field_types):
        return tuple(
            map(
                lambda f: f.name,
                filter(
                    lambda f: isinstance(f, field_types),
                    view.get_serializer_class().Meta.model()._meta.fields
                )
            )
        )

    def filter_by_func(self, queryset, queryset_func_name, related):
        if related:
            return getattr(queryset, queryset_func_name)(*related)
        return queryset

    def prefetch(self, queryset_func_name, view, queryset):
        with raise_context():
            queryset = self.filter_by_func(
                queryset,
                f'{queryset_func_name}_related',
                self.filter_model_fields(view, self.fields_fetch_map[queryset_func_name])
            )
        return queryset

    def filter_queryset(self, request, queryset, view):
        """
        Select related in queryset.
        """
        if queryset.query.select_related:
            return queryset
        if not getattr(view, 'select_related', True):
            return queryset
        if request.method != 'GET':
            return queryset
        return self.prefetch('prefetch', view, self.prefetch('select', view, queryset))


class DeepViewFilterBackend(VSTFilterBackend):
    """
    Backend that filters queryset by column from `deep_parent_field` property of the model.
    Value for filtering must be provided in query param `__deep_parent`.

    If param is missing then no filtering is applied.

    If param is empty value (`/?__deep_parent=`) then objects with no parent (the value of the field whose name is
    stored in the property `deep_parent_field` of the model is None) returned.

    This filter backend and nested view is automatically added when model has `deep_parent_field` property.

    Example:
        .. sourcecode:: python

            from django.db import models
            from vstutils.models import BModel

            class DeepNestedModel(BModel):
                name = models.CharField(max_length=10)
                parent = models.ForeignKey('self', null=True, default=None, on_delete=models.CASCADE)

                deep_parent_field = 'parent'
                deep_parent_allow_append = True

                class Meta:
                    default_related_name = 'deepnested'

    In example above if we add this model under path '`deep`', following views will be created: `/deep/` and
    `/deep/{id}/deepnested/`.

    Filter backend can be used as `/deep/?__deep_parent=1` and will return all `DeepNestedModel` objects
    whose parent's primary key is `1`.

    You can also use generic DRF views, for that you still must set `deep_parent_field`
    to your model and manually add `DeepViewFilterBackend` to
    `filter_backends <https://www.django-rest-framework.org/api-guide/filtering/#djangofilterbackend>`_ list.
    """
    field_name = '__deep_parent'
    field_types = {
        compat.coreschema.Integer: filters.NumberFilter,  # type: ignore
        compat.coreschema.String: filters.CharFilter,  # type: ignore
    }

    def filter_queryset(self, request, queryset, view):
        if view.action != 'list':
            return queryset
        model = queryset.model
        parent_name: _t.Optional[_t.Text] = getattr(model, 'deep_parent_field', None)
        if not parent_name or self.field_name not in request.query_params:
            return queryset
        filter_type_class = self.field_types[type(self.get_coreschema_field(model))]
        filter_data = request.query_params.get(self.field_name)
        if not filter_data:
            return queryset.filter(**{f'{parent_name}__isnull': True})
        pk_name = model._meta.pk.attname
        parent_qs = model.objects.filter(**{pk_name: filter_data})
        parent_qs = getattr(parent_qs, 'cleared', parent_qs.all)()
        return filter_type_class(field_name=parent_name, lookup_expr='in').filter(
            queryset,
            value=parent_qs.values(pk_name)
        )

    def get_coreschema_field(self, model: models.Model):
        primary_key_field: _t.Optional[models.Field] = model._meta.pk
        if isinstance(primary_key_field, models.IntegerField):
            field_cls = compat.coreschema.Integer  # type: ignore
        else:  # nocv
            field_cls = compat.coreschema.String  # type: ignore
        return field_cls(
            description=''
        )

    def get_schema_fields(self, view):
        return [
            compat.coreapi.Field(
                name=self.field_name,
                required=False,
                location='query',
                schema=self.get_coreschema_field(view.get_queryset().model),
            )
        ]



# pylint: disable=no-member,no-classmethod-decorator,protected-access
from django.db import models
from django.utils.functional import cached_property

from ..utils import Paginator, raise_context_decorator_with_default, is_member_descriptor


class BQuerySet(models.QuerySet):
    """
    Represent a lazy database lookup for a set of objects.
    Allows to override default iterable class by `custom_iterable_class` attribute
    (class with `__iter__` method which returns generator of model objects) and
    default query class by `custom_query_class` attribute
    (class inherited from :class:`django.db.models.sql.query.Query`).
    """

    __slots__ = ('__iterable_class__',)
    use_for_related_fields = True
    custom_query_class = None

    def __init__(self, model=None, query=None, using=None, hints=None):
        if query is None and self.custom_query_class is not None:
            query = self.custom_query_class(self)  # pylint: disable=not-callable
        super().__init__(model=model, query=query, using=using, hints=hints)

    @property
    def _iterable_class(self):
        if hasattr(self, '__iterable_class__') and not is_member_descriptor(self.__iterable_class__):
            return self.__iterable_class__
        if hasattr(self, 'custom_iterable_class'):  # pragma: no branch
            self.__iterable_class__ = self.custom_iterable_class
        return self._iterable_class

    @_iterable_class.setter
    def _iterable_class(self, value):
        if not hasattr(self, 'custom_iterable_class'):
            self.__iterable_class__ = value

    @_iterable_class.deleter
    def _iterable_class(self):  # nocv
        del self.__iterable_class__

    def paged(self, *args, **kwargs):
        """
        Returns paginated data with custom Paginator-class.
        Uses `PAGE_LIMIT` from global settings by default.
        """
        return self.get_paginator(*args, **kwargs).items()

    def get_paginator(self, *args, **kwargs):
        """
        Returns initialized object of :class:`vstutils.utils.Paginator` over
        current instance's QuerySet. All args and kwargs go to to Paginator's constructor.
        """
        return Paginator(self.filter(), *args, **kwargs)

    def has_field_filter_in_query(self, field_name):
        return any(filter(
            raise_context_decorator_with_default(default=False)(
                lambda x: getattr(x.lhs.field, 'attname', None) == field_name
            ),
            self.query.where.children
        ))

    @cached_property
    def has_hidden_filter(self):
        return self.has_field_filter_in_query('hidden')

    def cleared(self):
        """
        Filter queryset for models with attribute 'hidden' and
        exclude all hidden objects.
        """
        if hasattr(self.model, "hidden") and not self.has_hidden_filter:
            return self.filter(hidden=False)
        return self



# flake8: noqa: E501
TRANSLATION = {
    'new': 'Новый',
    'Create': 'создать',
    'add': 'добавить',
    'Edit': 'редактировать',
    'edit': 'Редактировать',
    'Save': 'Сохранить',
    'Remove': 'удалить',
    'Reload': 'перезагрузить',
    'send': 'отправить',
    'select': 'выбрать',
    'filters': 'фильтры',
    'apply': 'применить',
    'cancel': 'отменить',
    'field': 'поле',
    'actions': 'действия',
    'sublinks': 'вложенные ссылки',
    'name': 'наименование',
    'type': 'тип',
    'owner': 'владелец',
    'notes': 'заметки',
    'copy': 'копировать',
    'error': 'ошибка',
    'home': 'главная',
    'login': 'войти',
    'logout': 'выйти',
    'profile': 'профиль',
    'settings': 'настройки',
    'Settings': 'Настройки',
    'version': 'версий | версия | версии',
    'cache': 'кэш',
    'documentation': 'документация',
    'request': 'запрос',
    'repository': 'репозиторий',
    'app info': 'инфо',
    'status': 'статус',
    'list is empty': 'список пуст',
    'child instances': 'вложенные объекты',
    'enter value': 'введите значение',
    'more info': 'подробнее',
    'search by': 'поиск по полю',

    # x-menu generic
    'system': 'Система',
    'users': 'Пользователи',
    'Is active': 'Активен',
    'Is staff': 'Модератор',

    # field's validation
    # empty
    'Field "<b>{0}</b>" is empty.': 'Поле "<b>{0}</b>" пустое.',
    # required
    'Field "<b>{0}</b>" is required.': 'Поле "<b>{0}</b>" обязательно для заполнения.',
    # minLength
    'Field "<b>{0}</b>" is too short.<br> Field length should not be shorter, than {1}.': 'Длина поля "<b>{0}</b>" слишком мала.<br>Длина поля должна быть не меньше, чем {1}.',
    # maxLength
    'Field "<b>{0}</b>" is too long. <br> Field length should not be longer, than {1}.': 'Длина поля "<b>{0}</b>" слишком велика.<br>Длина поля должна быть не больше, чем {1}.',
    # min
    'Field "<b>{0}</b>" is too small.<br> Field should not be smaller, than {1}.': 'Значение поля "<b>{0}</b>" слишком мало.<br> Значение поля должно быть не меньше, чем {1}.',
    # max
    'Field "<b>{0}</b>" is too big.<br> Field should not be bigger, than {1}.': 'Значение поля "<b>{0}</b>" слишком велико.<br> Значение поля должно быть не больше, чем {1}.',
    # invalid
    '<b>{0} </b> value is not valid for <b>{1}</b> field.': 'Значение <b>{0}</b> не допустимо для поля <b>{1}</b>.',
    # email
    '<b>"{0}"</b> field should be written in <b>"example@mail.com"</b> format.': 'Поле <b>"{0}"</b> должно быть заполнено в формате <b>"example@mail.com"</b>.',


    # instance operation success
    # add
    'Child "<b>{0}</b>" instance was successfully added to parent list.': 'Дочерний объект "<b>{0}</b>" был успешно добавлен в родительский список.',
    # create
    'New "<b>{0}</b>" instance was successfully created.': 'Новый объект "<b>{0}</b>" был успешно создан.',
    # remove
    '"<b>{0}</b>" {1} was successfully removed.': '"<b>{0}</b>" {1} был(а) успешно удален.',
    # save
    'Changes in "<b>{0}</b>" {1} were successfully saved.': 'Изменения в объекте {1} "<b>{0}</b>" были успешно сохранены.',
    # execute
    'Action "<b>{0}</b>" was successfully executed on "<b>{1}</b>" instance.': 'Действие "<b>{0}</b>" было успешно запущено на объекте "<b>{1}</b>".',

    # instance operation error
    # add:
    'Some error occurred during adding of child "<b>{0}</b>" instance to parent list.<br> Error details: {1}': 'Во время добавления дочернего объекта "<b>{0}</b>" к родительскому списку произошла ошибка.<br> Подробнее: {1}',
    # create:
    'Some error occurred during new "<b>{0}</b>" instance creation.<br> Error details: {1}': 'Во время создания нового объекта "<b>{0}</b>" произошла ошибка.<br> Подробнее: {1}',
    # remove:
    'Some error occurred during remove process of "<b>{0}</b>" {1}.<br> Error details: {2}': 'Во время удаления {1} "<b>{0}</b>" произошла ошибка.<br> Подробнее: {2}',
    # save:
    'Some error occurred during saving process of "<b>{0}</b>" {1}.<br> Error details: {2}': 'Во время сохранения {1} "<b>{0}</b>" произошла ошибка.<br> Подробнее: {2}',
    # execute:
    'Some error occurred during "<b>{0}</b>" action execution on {1}.<br> Error details: {2}': 'Во время запуска действия "<b>{0}</b>" на {1} произошла ошибка.<br> Подробнее: {2}',

    # guiCustomizer translations
    'customize application styles': 'настройка цветовых стилей приложения',
    'skin': 'тема',
    'skin settings': 'настройки темы',
    'active menu background': 'фон активного элемента меню',
    'active menu color': 'активный элемент меню',
    'body background': 'фон страницы',
    'top navigation background': 'фон верхнего блока навигации',
    'top navigation border': 'окантовка верхнего блока навигации',
    'left sidebar background': 'фон сайдбара слева',
    'left sidebar border': 'окантовка сайдбара слева',
    'customizer sidebar background': 'фон данного сайдбара',
    'breadcrumbs background': 'фон хлебный крошек',
    'buttons default background': 'фон стандартной кнопки',
    'buttons default text': 'текст стандартной кнопки',
    'buttons default border': 'окантовка стандартной кнопки',
    'buttons primary background': 'фон основной кнопки',
    'buttons primary text': 'текст основной кнопки',
    'buttons primary border': 'окантовка основной кнопки',
    'buttons danger background': 'фон danger кнопки',
    'buttons danger text': 'текст danger кнопки',
    'buttons danger border': 'окантовка danger кнопки',
    'buttons warning background': 'фон warning кнопки',
    'buttons warning text': 'текст warning кнопки',
    'buttons warning border': 'окантовка warning кнопки',
    'links': 'ссылка',
    'links hover': 'ссылка при наведении',
    'links revers': 'ссылка на темном фоне',
    'links hover revers': 'ссылка на темном фоне при наведении',
    'text color': 'текст',
    'ico default color': 'стандартаный цвет иконки',
    'text header color': 'заголовок',
    'background default color': 'стандартный цвет фона',
    'card header background': 'фон верхней части "card" блока',
    'card body background': 'фон основной части "card" блока',
    'card footer background': 'фон нижней части "card" блока',
    'card color': 'текст "card" в блоке',
    'labels': 'лейбл',
    'help content': 'вспомогательный контент',
    'help text color': 'вспомогательный текст',
    'table line hover bg color': 'фон табличного ряда при наведении',
    'table border color': 'окантовка таблицы',
    'table line selected bg color': 'фон выделенного табличного ряда',
    'background active color': 'фон активного элемента',
    'background passive color': 'фон неактивного элемента',
    'text hover color': 'цвет текста при наведении',
    'boolean true color': 'цвет для булевых переменных со значением "true"',
    'boolean false color': 'цвет для булевых переменных со значением "false"',
    'modal background color': 'фон модального окна',
    'api sections background color': 'фон секций в Api',
    'api sections border color': 'окантовка секций в Api',
    'code highlighting color 1': 'подсветка кода - цвет 1',
    'code highlighting color 2': 'подсветка кода - цвет 2',
    'code highlighting color 3': 'подсветка кода - цвет 3',
    'code highlighting color 4': 'подсветка кода - цвет 4',
    'code highlighting color 5': 'подсветка кода - цвет 5',
    'code highlighting color 6': 'подсветка кода - цвет 6',
    'code highlighting color 7': 'подсветка кода - цвет 7',

    'custom css': 'пользовательский css',
    'reset skin settings to default': 'к настройкам по умолчанию',
    'save skin': 'сохранить тему',
    'Skin settings were successfully reset to default.': 'Настройки темы были успешно возвращены к стандратным.',
    'Skin settings were successfully saved': 'Настройки темы были успешно сохранены.',

    # multiactions button
    'actions on': 'действия над',
    'on item': 'элементов | элементом | элементами | элементами',

    'key': 'ключ',
    'value': 'значение',
    'new owner': 'новый владелец',
    'detail': 'детали',
    'description': 'описание',
    'username': 'имя пользователя',
    'is active': 'активен',
    'user': 'пользователь',
    'yes': 'да',
    'no': 'нет',
    'is staff': 'админ',
    'first name': 'имя',
    'last name': 'фамилия',
    'password': 'пароль',
    'Password': 'пароль',
    'Change password': 'сменить пароль',
    'Change_password': 'сменить пароль',
    'generate password': 'сформировать пароль',
    'repeat password': 'повторить пароль',
    'Repeat password': 'повторить пароль',
    'Old password': 'старый пароль',
    'New password': 'новый пароль',
    'Confirm new password': 'подтвердить новый пароль',
    'kind': 'вид',
    'mode': 'режим',
    'options': 'опции',
    'information': 'информация',
    'date': 'дата',

    'file n selected': 'файл не выбран | 1 файл выбран | {n} файла выбрано | {n} файлов выбрано',
    'image n selected': 'изображение не выбрано | 1 изображение выбрано | {n} изображения выбрано | {n} изображений выбрано',

    # filters
    'Ordering': 'Порядок сортировки',
    'is_active': 'активен',
    'Username': 'Ник',
    'First name': 'Имя',
    'Last name': 'Фамилия',

    # filters description
    'boolean value meaning status of user.': 'булево значение обозначающее статус пользователя.',
    'users first name.': 'имя пользователя',
    'users last name.': 'фамилия пользователя',
    'users e-mail value.': 'email пользователя',
    'which field to use when ordering the results.': 'поле по которому производить сортировку результатов.',
    'a unique integer value (or comma separated list) identifying this instance.': 'уникальное числовое значение (или их последовательность разделенная запятой) идентифицирующая данный объект.',
    'a name string value (or comma separated list) of instance.': 'имя объекта - строковое значение (или их последовательность разделенная запятой).',

    'page matching current url was not found': 'страница, соответствующая данному url, не была найдена',

    'sign in to start your session': 'Войдите, чтобы начать сеанс',
    'sign in': 'войти',
    'sign up': 'зарегистрируйтесь',
    'forgot password': 'забыли пароль',
    'confirm password': 'подтвердите пароль',
    'register': 'регистрация',
    'fields with * is required': 'поля отмеченые * обязательны для заполнения',
    'Forgotten your password? Enter your email address below, and we\'ll email instructions for setting a new one.': 'Забыли пароль? Введите свой адрес и мы отправим вам инструкцию для восстановления.',
    'reset my password': 'сбросить пароль',
    'We\'ve emailed you instructions for setting your password, if an account exists with the email you entered. You should receive them shortly.': 'Мы отправили вам письмо с инструкцией для сброса пароля.',
    'If you don\'t receive an email, please make sure you\'ve entered the address you registered with, and check your spam folder.': 'Если вы не получили письмо, пожалуйста убедитесь что введен верный адрес и проверьте папку со спамом.',
    'The password reset link was invalid, possibly because it has already been used. Please request a new password reset.': 'Неверная ссылка для восстановления пароля, возможно она уже была использована. Пожалуйста попробуйте заново.',
    'Please enter your new password twice so we can verify you typed it in correctly.': 'Пожалуйста введите новый пароль.',
    'change my password': 'изменить пароль',
    'Your password has been set.  You may go ahead and log in now.': 'Ваш пароль был успешно установлен. Теперь вы можете войти.',
    'Please enter a correct %(username)s and password. Note that both fields may be case-sensitive.': 'Пожалуйста введите корректные %(username)s и пароль. Внимание, поля могуь быть регистрозвисимыми.',
    'or': 'или',
    'two-factor authentication code': 'код двухфакторной авторизации',
    'verify code': 'подтвердить код',
    'Enter the code from the two-factor app on your mobile device. If you\'ve lost your device, you may enter one of your recovery codes.': 'Введите код из двухфакторного приложения на мобильном устройстве. Если вы потеряли свое устройство, вы можете ввести один из кодов восстановления.',
    'Invalid authentication code': 'Неверный код авторизации',
    'Scan the image with the two-factor authentication app on your phone. If you can’t use a barcode, enter this text code instead.': 'Отсканируйте изображение с помощью приложения для двухфакторной аутентификации на вашем телефоне. Если вы не можете использовать штрих-код, введите вместо него текстовый код.',
    'Enter the six-digit code from the application': 'Введите шестисимвольный код из приложения',
    'code': 'код',
    'recovery codes': 'коды восстановления',
    'Recovery codes are used to access your account in the event you cannot receive two-factor authentication codes.': 'Коды восстановления используются для доступа к вашей учетной записи в случае, если вы не можете получить коды двухфакторной аутентификации.',
    'Disabling Two-Factor Authentication (2FA) will decrease your account\'s security': 'Отключение двухфакторной авторизации снизит безопасность вашей учетной записи.',
    'Enable': 'Включить',
    'Disable': 'Выключить',
    'Two factor authentication': 'Двухфакторная авторизация',
    'No {0} matches the given query.': "Ни один {0} не соответствует данному запросу.",
    'Password is not correct.': 'Неверный пароль.',
    'New passwords values are not equal.': 'Значения нового пароля не эквивалентны.',
    'This password is entirely numeric.': 'Этот пароль полностью числовой.',
    'Dark theme': 'Темная тема',
    'Language': 'Язык',
    'height': 'высота',
    'width': 'ширина',
    'More info': 'Подробнее',
    'Dark mode': 'Темная тема',

    # validators
    'unsupported image file format, expected ({0}), got {1}': 'Неподдерживаемый формат изображение, Поддерживается ({0}), получен {1}',
    'for some reason, this image file cannot be opened': 'Не удалось открыть файл изображения',
    'Invalid image {0}. Expected from {1} to {2}, got {3}': 'Некорректная {0} изображения. Ожидалось от {1} до {2}, получено {3}',
    'imageValidationResolutionError': 'Должно быть не меньше {min} и не больше {max} px.',

    # Проверить
    'Changes in settings are successfully saved. Please refresh the page.': 'Изменения в настройках успешно сохранены. Пожалуйста обновите страницу.',
    'now': 'сейчас',
    'later': 'позже',
    'i accept the ': 'Я принимаю ',
    'terms of agreement': 'лицензионное соглашение',
    'Zoom': 'Зум',
    'Scale': 'Масштаб',
}



# Stubs for vstutils.api.responses (Python 3)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from rest_framework import response, status
from typing import Any, Dict, Text, SupportsFloat, Optional

class BaseResponseClass(response.Response):
    data: Any
    timings: Optional[Dict[Text, SupportsFloat]]
    def __init__(self, *args: Any, **kwargs: Any) -> None: ...


class HTTP_100_CONTINUE(BaseResponseClass):
    status_code = status.HTTP_100_CONTINUE


class Response100(HTTP_100_CONTINUE):...


class HTTP_101_SWITCHING_PROTOCOLS(BaseResponseClass):
    status_code = status.HTTP_101_SWITCHING_PROTOCOLS


class Response101_SWITCHING(HTTP_101_SWITCHING_PROTOCOLS):...


class HTTP_200_OK(BaseResponseClass):
    status_code = status.HTTP_200_OK


class Response200(HTTP_200_OK):...


class HTTP_201_CREATED(BaseResponseClass):
    status_code = status.HTTP_201_CREATED


class Response201(HTTP_201_CREATED):...


class HTTP_202_ACCEPTED(BaseResponseClass):
    status_code = status.HTTP_202_ACCEPTED


class Response202(HTTP_202_ACCEPTED):...


class HTTP_203_NON_AUTHORITATIVE_INFORMATION(BaseResponseClass):
    status_code = status.HTTP_203_NON_AUTHORITATIVE_INFORMATION


class Response203_NON_AUTHORITATIVE(HTTP_203_NON_AUTHORITATIVE_INFORMATION):...


class HTTP_204_NO_CONTENT(BaseResponseClass):
    status_code = status.HTTP_204_NO_CONTENT


class Response204_NO(HTTP_204_NO_CONTENT):...


class HTTP_205_RESET_CONTENT(BaseResponseClass):
    status_code = status.HTTP_205_RESET_CONTENT


class Response205_RESET(HTTP_205_RESET_CONTENT):...


class HTTP_206_PARTIAL_CONTENT(BaseResponseClass):
    status_code = status.HTTP_206_PARTIAL_CONTENT


class Response206_PARTIAL(HTTP_206_PARTIAL_CONTENT):...


class HTTP_207_MULTI_STATUS(BaseResponseClass):
    status_code = status.HTTP_207_MULTI_STATUS


class Response207_MULTI(HTTP_207_MULTI_STATUS):...


class HTTP_300_MULTIPLE_CHOICES(BaseResponseClass):
    status_code = status.HTTP_300_MULTIPLE_CHOICES


class Response300_MULTIPLE(HTTP_300_MULTIPLE_CHOICES):...


class HTTP_301_MOVED_PERMANENTLY(BaseResponseClass):
    status_code = status.HTTP_301_MOVED_PERMANENTLY


class Response301_MOVED(HTTP_301_MOVED_PERMANENTLY):...


class HTTP_302_FOUND(BaseResponseClass):
    status_code = status.HTTP_302_FOUND


class Response302(HTTP_302_FOUND):...


class HTTP_303_SEE_OTHER(BaseResponseClass):
    status_code = status.HTTP_303_SEE_OTHER


class Response303_SEE(HTTP_303_SEE_OTHER):...


class HTTP_304_NOT_MODIFIED(BaseResponseClass):
    status_code = status.HTTP_304_NOT_MODIFIED


class Response304_NOT(HTTP_304_NOT_MODIFIED):...


class HTTP_305_USE_PROXY(BaseResponseClass):
    status_code = status.HTTP_305_USE_PROXY


class Response305_USE(HTTP_305_USE_PROXY):...


class HTTP_306_RESERVED(BaseResponseClass):
    status_code = status.HTTP_306_RESERVED


class Response306(HTTP_306_RESERVED):...


class HTTP_307_TEMPORARY_REDIRECT(BaseResponseClass):
    status_code = status.HTTP_307_TEMPORARY_REDIRECT


class Response307_TEMPORARY(HTTP_307_TEMPORARY_REDIRECT):...


class HTTP_400_BAD_REQUEST(BaseResponseClass):
    status_code = status.HTTP_400_BAD_REQUEST


class Response400_BAD(HTTP_400_BAD_REQUEST):...


class HTTP_401_UNAUTHORIZED(BaseResponseClass):
    status_code = status.HTTP_401_UNAUTHORIZED


class Response401(HTTP_401_UNAUTHORIZED):...


class HTTP_402_PAYMENT_REQUIRED(BaseResponseClass):
    status_code = status.HTTP_402_PAYMENT_REQUIRED


class Response402_PAYMENT(HTTP_402_PAYMENT_REQUIRED):...


class HTTP_403_FORBIDDEN(BaseResponseClass):
    status_code = status.HTTP_403_FORBIDDEN


class Response403(HTTP_403_FORBIDDEN):...


class HTTP_404_NOT_FOUND(BaseResponseClass):
    status_code = status.HTTP_404_NOT_FOUND


class Response404_NOT(HTTP_404_NOT_FOUND):...


class HTTP_405_METHOD_NOT_ALLOWED(BaseResponseClass):
    status_code = status.HTTP_405_METHOD_NOT_ALLOWED


class Response405_METHOD_NOT(HTTP_405_METHOD_NOT_ALLOWED):...


class HTTP_406_NOT_ACCEPTABLE(BaseResponseClass):
    status_code = status.HTTP_406_NOT_ACCEPTABLE


class Response406_NOT(HTTP_406_NOT_ACCEPTABLE):...


class HTTP_407_PROXY_AUTHENTICATION_REQUIRED(BaseResponseClass):
    status_code = status.HTTP_407_PROXY_AUTHENTICATION_REQUIRED


class Response407_PROXY_AUTHENTICATION(HTTP_407_PROXY_AUTHENTICATION_REQUIRED):...


class HTTP_408_REQUEST_TIMEOUT(BaseResponseClass):
    status_code = status.HTTP_408_REQUEST_TIMEOUT


class Response408_REQUEST(HTTP_408_REQUEST_TIMEOUT):...


class HTTP_409_CONFLICT(BaseResponseClass):
    status_code = status.HTTP_409_CONFLICT


class Response409(HTTP_409_CONFLICT):...


class HTTP_410_GONE(BaseResponseClass):
    status_code = status.HTTP_410_GONE


class Response410(HTTP_410_GONE):...


class HTTP_411_LENGTH_REQUIRED(BaseResponseClass):
    status_code = status.HTTP_411_LENGTH_REQUIRED


class Response411_LENGTH(HTTP_411_LENGTH_REQUIRED):...


class HTTP_412_PRECONDITION_FAILED(BaseResponseClass):
    status_code = status.HTTP_412_PRECONDITION_FAILED


class Response412_PRECONDITION(HTTP_412_PRECONDITION_FAILED):...


class HTTP_413_REQUEST_ENTITY_TOO_LARGE(BaseResponseClass):
    status_code = status.HTTP_413_REQUEST_ENTITY_TOO_LARGE


class Response413_REQUEST_ENTITY_TOO(HTTP_413_REQUEST_ENTITY_TOO_LARGE):...


class HTTP_414_REQUEST_URI_TOO_LONG(BaseResponseClass):
    status_code = status.HTTP_414_REQUEST_URI_TOO_LONG


class Response414_REQUEST_URI_TOO(HTTP_414_REQUEST_URI_TOO_LONG):...


class HTTP_415_UNSUPPORTED_MEDIA_TYPE(BaseResponseClass):
    status_code = status.HTTP_415_UNSUPPORTED_MEDIA_TYPE


class Response415_UNSUPPORTED_MEDIA(HTTP_415_UNSUPPORTED_MEDIA_TYPE):...


class HTTP_416_REQUESTED_RANGE_NOT_SATISFIABLE(BaseResponseClass):
    status_code = status.HTTP_416_REQUESTED_RANGE_NOT_SATISFIABLE


class Response416_REQUESTED_RANGE_NOT(HTTP_416_REQUESTED_RANGE_NOT_SATISFIABLE):...


class HTTP_417_EXPECTATION_FAILED(BaseResponseClass):
    status_code = status.HTTP_417_EXPECTATION_FAILED


class Response417_EXPECTATION(HTTP_417_EXPECTATION_FAILED):...


class HTTP_422_UNPROCESSABLE_ENTITY(BaseResponseClass):
    status_code = status.HTTP_422_UNPROCESSABLE_ENTITY


class Response422_UNPROCESSABLE(HTTP_422_UNPROCESSABLE_ENTITY):...


class HTTP_423_LOCKED(BaseResponseClass):
    status_code = status.HTTP_423_LOCKED


class Response423(HTTP_423_LOCKED):...


class HTTP_424_FAILED_DEPENDENCY(BaseResponseClass):
    status_code = status.HTTP_424_FAILED_DEPENDENCY


class Response424_FAILED(HTTP_424_FAILED_DEPENDENCY):...


class HTTP_428_PRECONDITION_REQUIRED(BaseResponseClass):
    status_code = status.HTTP_428_PRECONDITION_REQUIRED


class Response428_PRECONDITION(HTTP_428_PRECONDITION_REQUIRED):...


class HTTP_429_TOO_MANY_REQUESTS(BaseResponseClass):
    status_code = status.HTTP_429_TOO_MANY_REQUESTS


class Response429_TOO_MANY(HTTP_429_TOO_MANY_REQUESTS):...


class HTTP_431_REQUEST_HEADER_FIELDS_TOO_LARGE(BaseResponseClass):
    status_code = status.HTTP_431_REQUEST_HEADER_FIELDS_TOO_LARGE


class Response431_REQUEST_HEADER_FIELDS_TOO(HTTP_431_REQUEST_HEADER_FIELDS_TOO_LARGE):...


class HTTP_451_UNAVAILABLE_FOR_LEGAL_REASONS(BaseResponseClass):
    status_code = status.HTTP_451_UNAVAILABLE_FOR_LEGAL_REASONS


class Response451_UNAVAILABLE_FOR_LEGAL(HTTP_451_UNAVAILABLE_FOR_LEGAL_REASONS):...


class HTTP_500_INTERNAL_SERVER_ERROR(BaseResponseClass):
    status_code = status.HTTP_500_INTERNAL_SERVER_ERROR


class Response500_INTERNAL_SERVER(HTTP_500_INTERNAL_SERVER_ERROR):...


class HTTP_501_NOT_IMPLEMENTED(BaseResponseClass):
    status_code = status.HTTP_501_NOT_IMPLEMENTED


class Response501_NOT(HTTP_501_NOT_IMPLEMENTED):...


class HTTP_502_BAD_GATEWAY(BaseResponseClass):
    status_code = status.HTTP_502_BAD_GATEWAY


class Response502_BAD(HTTP_502_BAD_GATEWAY):...


class HTTP_503_SERVICE_UNAVAILABLE(BaseResponseClass):
    status_code = status.HTTP_503_SERVICE_UNAVAILABLE


class Response503_SERVICE(HTTP_503_SERVICE_UNAVAILABLE):...


class HTTP_504_GATEWAY_TIMEOUT(BaseResponseClass):
    status_code = status.HTTP_504_GATEWAY_TIMEOUT


class Response504_GATEWAY(HTTP_504_GATEWAY_TIMEOUT):...


class HTTP_505_HTTP_VERSION_NOT_SUPPORTED(BaseResponseClass):
    status_code = status.HTTP_505_HTTP_VERSION_NOT_SUPPORTED


class Response505_HTTP_VERSION_NOT(HTTP_505_HTTP_VERSION_NOT_SUPPORTED):...


class HTTP_507_INSUFFICIENT_STORAGE(BaseResponseClass):
    status_code = status.HTTP_507_INSUFFICIENT_STORAGE


class Response507_INSUFFICIENT(HTTP_507_INSUFFICIENT_STORAGE):...


class HTTP_511_NETWORK_AUTHENTICATION_REQUIRED(BaseResponseClass):
    status_code = status.HTTP_511_NETWORK_AUTHENTICATION_REQUIRED


class Response511_NETWORK_AUTHENTICATION(HTTP_511_NETWORK_AUTHENTICATION_REQUIRED):...


